from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from dsp_tools.error.problems import Problem

separator = "\n    "
list_separator = "\n    - "
medium_separator = "\n----------------------------\n"

grand_separator = "\n\n---------------------------------------\n\n"


@dataclass(frozen=True)
class PositionInExcel:
    """This class contains the information about the position of a value in the excel."""

    sheet: str | None = None
    column: str | None = None
    row: int | None = None
    excel_filename: str | None = None

    def __str__(self) -> str:
        msg = []
        if self.excel_filename:
            msg.append(f"Excel '{self.excel_filename}'")
        if self.sheet:
            msg.append(f"Sheet '{self.sheet}'")
        if self.column:
            msg.append(f"Column '{self.column}'")
        if self.row:
            msg.append(f"Row {self.row}")
        return " | ".join(msg)

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return str(self)


@dataclass(frozen=True)
class ExcelFileProblem:
    """This class contains information about problems with an excel file."""

    filename: str
    problems: list[Problem]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        msg = f"The Excel file '{self.filename}' contains the following problems:\n\n"
        msg += medium_separator.join([x.execute_error_protocol() for x in self.problems])
        return msg


@dataclass(frozen=True)
class ExcelSheetProblem:
    """This class contains information about problems in one excel sheet."""

    sheet_name: str
    problems: list[Problem]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """

        problem_strings = [x.execute_error_protocol() for x in self.problems]
        return f"The sheet '{self.sheet_name}' has the following problems:\n\n{'\n\n'.join(problem_strings)}"


@dataclass(frozen=True)
class EmptySheetsProblem:
    """This class contains information if sheets that must contain information are empty."""

    sheet_names: list[str]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"The following sheet(s) are mandatory and may not be empty:{list_separator}"
            f"{list_separator.join(sorted(self.sheet_names))}"
        )


@dataclass(frozen=True)
class MissingValuesProblem:
    """This class contains information if a value is missing in a location"""

    locations: list[PositionInExcel]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        locs = [str(x) for x in self.locations]
        return f"At the following locations, mandatory values are missing:{list_separator}{list_separator.join(locs)}"


@dataclass(frozen=True)
class RequiredColumnMissingProblem:
    """This class contains information if a required column is missing."""

    columns: list[str]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return f"The following required column(s) are missing:{list_separator}{list_separator.join(self.columns)}"


@dataclass(frozen=True)
class DuplicatesInColumnProblem:
    """This class contains information if a required column is missing."""

    column: str
    duplicate_values: list[str]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"No duplicates are allowed in the column '{self.column}'\n"
            f"The following values appear several times:{list_separator}"
            f"{list_separator.join(self.duplicate_values)}"
        )


@dataclass(frozen=True)
class InvalidExcelContentProblem:
    """This class contains information if a required column is missing."""

    expected_content: str
    actual_content: str
    excel_position: PositionInExcel

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"There is invalid content in the excel.\n"
            f"Located at: {self.excel_position}\n"
            f"Expected Content: {self.expected_content}\n"
            f"Actual Content: {self.actual_content}"
        )


@dataclass(frozen=True)
class InvalidSheetNameProblem:
    """This class contains information if the excel sheet names are not strings."""

    excelfile: str
    excel_sheet_names: list[Any]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        sheet_types = [f"Name: {x} | Type {type(x)}" for x in self.excel_sheet_names if not isinstance(x, str)]
        return (
            f"The names sheets in the excel '{self.excelfile}' are not all valid.\n"
            f"They must be of type string. The following names are problematic:\n"
            f"{list_separator}{list_separator.join(sheet_types)}\n"
            f"Please rename them."
        )


@dataclass(frozen=True)
class DuplicateSheetProblem:
    duplicate_sheets: list[str]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"The sheet names inside the same Excel file must be unique. "
            f"Using capitalisation or spaces to differentiate sheets is not valid.\n"
            f"For example 'sheet' and 'SHEET  ' are considered identical.\n"
            f"Under this condition, the following sheet names appear multiple times:"
            f"{list_separator}{list_separator.join(sorted(self.duplicate_sheets))}"
        )


@dataclass(frozen=True)
class MoreThanOneSheetProblem:
    """This class contains information if the excel containing the property values has more than one sheet."""

    excelname: str
    sheet_names: list[str]

    def execute_error_protocol(self) -> str:
        msg = [
            f"\nIn the '{self.excelname}' file only one sheet is allowed.",
            f"The excel used contains the following sheets:{list_separator}{list_separator.join(self.sheet_names)}",
            "Please delete all but one sheet.",
        ]
        return separator.join(msg)


@dataclass(frozen=True)
class MandatorySheetsMissingProblem:
    """This class contains information if the excel is missing a mandatory sheet."""

    mandatory_sheet: list[str]
    existing_sheets: list[str]

    def execute_error_protocol(self) -> str:
        return (
            f"The following mandatory sheet(s) are missing in this Excel:{list_separator}"
            f"{list_separator.join(sorted(self.mandatory_sheet))}"
            f"The following sheets are in the file:{list_separator}"
            f"{list_separator.join(sorted(self.existing_sheets))}"
        )


@dataclass(frozen=True)
class ResourceSheetNotListedProblem:
    """This class contains information if some resource sheets are missing in the 'classes' sheet."""

    missing_names: set[str]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"All the sheets in the Excel must be listed in the 'name' column of the 'classes' sheet.\n"
            f"The following sheet(s) are not listed in the column:{list_separator}"
            f"{list_separator.join(self.missing_names)}"
        )


@dataclass(frozen=True)
class MoreThanOneRowProblem:
    """This class contains information if a sheet has more than one row."""

    num_rows: int

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return f"This sheet may only have one row. The provided sheet has {self.num_rows} rows."


@dataclass(frozen=True)
class AtLeastOneValueRequiredProblem:
    """This class contains information if at least one entry in a group of columns must be filled."""

    columns: list[str]
    row_num: int

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        return (
            f"At least one value is required in the columns: {', '.join(sorted(self.columns))}\n"
            f"Row {self.row_num} does not contain any values in those columns."
        )


@dataclass(frozen=True)
class JsonValidationPropertyProblem:
    """This class contains information about a JSON property section that fails its validation against the schema."""

    problematic_property: str | None = None
    original_msg: str | None = None
    message_path: str | None = None
    excel_position: PositionInExcel | None = None

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        msg = [
            f"{separator}Section of the problem: 'Properties'",
        ]
        if self.problematic_property:
            msg.append(f"Problematic property: '{self.problematic_property}'")
        if self.excel_position:
            msg.append(f"Located at: {self.excel_position}")
        if self.original_msg:
            msg.append(f"Original Error Message: {self.original_msg}")
        if self.message_path:
            msg.append(f"The error occurred at {self.message_path}")
        return separator.join(msg)


@dataclass(frozen=True)
class JsonValidationResourceProblem:
    """This class contains information about a JSON resource section that fails its validation against the schema."""

    problematic_resource: str | None = None
    excel_position: PositionInExcel | None = None
    original_msg: str | None = None
    message_path: str | None = None

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        msg = [
            f"{separator}Section of the problem: 'Resources'",
        ]
        if self.problematic_resource:
            msg.append(f"Problematic Resource '{self.problematic_resource}'")
        if self.excel_position:
            msg.append(f"Located at: {self.excel_position}")
        if self.original_msg:
            msg.append(f"Original Error Message: {self.original_msg}")
        if self.message_path:
            msg.append(f"The error occurred at {self.message_path}")
        return separator.join(msg)


@dataclass(frozen=True)
class PropertyProblem:
    """This class contains information if a property has invalid content."""

    prop_name: str
    problems: list[Problem]

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.

        Returns:
            message for the error
        """
        all_problems = [x.execute_error_protocol() for x in self.problems]
        return f"The property '{self.prop_name}' has the following problem(s):\n{medium_separator.join(all_problems)}"



from dataclasses import dataclass

import pandas as pd

list_separator = "\n    - "
medium_separator = "\n----------------------------\n"
grand_separator = "\n\n---------------------------------------\n\n"

maximum_prints = 50


@dataclass
class TextValueData:
    resource_id: str
    property_name: str
    encoding: set[str]


@dataclass
class InconsistentTextValueEncodings:
    """
    This class implements the `Problem` protocol
    for resources and properties that contain invalid encodings.

    An invalid encoding would be a `<text-prop>` element, that contains
    `<text encoding="utf8">`
    and
    `<text encoding="xml">`
    """

    problematic_resources: list[TextValueData]

    def execute_problem_protocol(self) -> tuple[str, pd.DataFrame | None]:
        """
        This method composes an error message for the user.
        If the number of errors exceeds `maximum_prints`,
        the errors are additionally returned as a dataframe that can be saved as a CSV file.

        Returns:
            the error message, and optionally a dataframe with the errors
        """
        base_msg = (
            "\nSome <text-prop> elements contain <text> elements that use both 'xml' and 'utf8' encoding.\n"
            "Only one encoding type can be used within one <text-prop> element."
        )
        df = self._get_problems_as_df()
        if len(df) > maximum_prints:
            return base_msg, df
        return base_msg + grand_separator + _make_msg_from_df(df), None

    def _get_problems_as_df(self) -> pd.DataFrame:
        df = pd.DataFrame(
            {
                "Resource ID": [x.resource_id for x in self.problematic_resources],
                "Property Name": [x.property_name for x in self.problematic_resources],
            }
        )
        return df.sort_values(by=["Resource ID", "Property Name"], ignore_index=True)


def _make_msg_from_df(df: pd.DataFrame) -> str:
    groups = df.groupby(by="Resource ID")
    return medium_separator.join([_make_msg_for_one_resource(str(_id), res_df) for _id, res_df in groups])


def _make_msg_for_one_resource(res_id: str, res_df: pd.DataFrame) -> str:
    problems = [f"Property Name: '{p}'" for p in res_df["Property Name"].tolist()]
    return f"Resource ID: '{res_id}'{list_separator}{list_separator.join(problems)}"



from pathlib import Path

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from dsp_tools.commands.ingest_xmlupload.create_resources.user_information import IngestInformation


class TestIngestInformation:
    def test_no_problems(self) -> None:
        expected = (
            "All multimedia files referenced in the XML file were uploaded through dsp-ingest.\n"
            "All multimedia files uploaded through dsp-ingest were referenced in the XML file."
        )
        assert IngestInformation([], []).ok_msg() == expected

    def test_not_all_good(self) -> None:
        assert not IngestInformation(["unused_media"], []).ok_msg()

    def test_unused_media(self) -> None:
        expected = (
            "The upload cannot continue as there are problems with the multimedia files referenced in the XML.\n"
            "    The data XML file does not reference the following multimedia files "
            "which were previously uploaded through dsp-ingest:\n"
            "    - unused_media"
        )
        res_info = IngestInformation(["unused_media"], [])._get_error_msg()
        assert res_info == expected

    def test_not_uploaded(self) -> None:
        expected = (
            "The upload cannot continue as there are problems with the multimedia files referenced in the XML.\n"
            "    The data XML file contains references to the following multimedia files "
            "which were not previously uploaded through dsp-ingest:\n"
            "    - Resource ID: 'no upload id' | Filepath: 'media path'"
        )
        res_info = IngestInformation([], [("no upload id", "media path")])._get_error_msg()
        assert res_info == expected

    def test_all_problem_with_df_msg(self) -> None:
        expected = (
            "The upload cannot continue as there are problems with the multimedia files referenced in the XML.\n"
            "    The data XML file does not reference all the multimedia files "
            "which were previously uploaded through dsp-ingest.\n"
            "    The file with the unused filenames was saved at 'UnusedMediaUploaded.csv'.\n"
            "    The data XML file contains references to multimedia files "
            "which were not previously uploaded through dsp-ingest:\n"
            "    The file with the resource IDs and problematic filenames was saved at 'FilesNotUploaded.csv'."
        )
        res_info = IngestInformation(
            [
                "unused_media0",
                "unused_media1",
            ],
            [
                ("no_up0_id", "no_up0_file"),
                ("no_up1_id", "no_up1_file"),
            ],
            csv_directory_path=Path(),
            maximum_prints=1,
        )._get_error_msg()
        assert res_info == expected


def test_unused_media_to_df() -> None:
    unused_media_list = [
        "unused_media0",
        "unused_media1",
    ]
    expected = pd.DataFrame({"Multimedia Filenames": unused_media_list})
    res_df = IngestInformation(unused_media_list, [], maximum_prints=1)._unused_mediafiles_to_df()
    assert res_df is not None
    assert_frame_equal(res_df, expected)


def test_unused_media_to_df_not_enough() -> None:
    assert not IngestInformation(["unused_media"], [], maximum_prints=1)._unused_mediafiles_to_df()


def test_no_id_to_df() -> None:
    expected = pd.DataFrame(
        {
            "Resource ID": [
                "ID_no_up0",
                "ID_no_up1",
            ],
            "Filepath": [
                "fileno_up0.jpg",
                "fileno_up1.jpg",
            ],
        }
    )
    res_df = IngestInformation(
        [],
        [
            ("ID_no_up0", "fileno_up0.jpg"),
            ("ID_no_up1", "fileno_up1.jpg"),
        ],
        maximum_prints=1,
    )._mediafiles_no_id_to_df()
    assert res_df is not None
    assert_frame_equal(res_df, expected)


def test_no_id_to_df_not_enough() -> None:
    assert not IngestInformation([], [("ID_no_up0", "fileno_up0.jpg")])._unused_mediafiles_to_df()


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="no-untyped-def"

import pytest
import regex

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import IIIFUriInfo
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLBitstream
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLFileMetadata
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLProperty
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileValue
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryIIIFUri
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.res import MigrationMetadata
from dsp_tools.commands.xmlupload.models.intermediary.res import ResourceInputConversionFailure
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryBoolean
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryColor
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDate
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDecimal
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeoname
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInt
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryList
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryTime
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryUri
from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import _get_metadata
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    _transform_file_value,
)
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    _transform_iiif_uri_value,
)
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    _transform_into_intermediary_resource,
)
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    _transform_one_property,
)
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    _transform_one_resource,
)
from dsp_tools.error.exceptions import InputError
from dsp_tools.error.exceptions import PermissionNotExistsError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.data_formats.date_util import Date

ONTO = "http://0.0.0.0:3333/ontology/9999/onto/v2#"


class TestTransformResources:
    def test_success(
        self,
        resource_one_prop: XMLResource,
        lookups: IntermediaryLookups,
    ):
        result = _transform_into_intermediary_resource(resource_one_prop, lookups)
        assert isinstance(result, IntermediaryResource)

    def test_failure(self, resource_with_unknown_permissions: XMLResource, lookups: IntermediaryLookups):
        result = _transform_into_intermediary_resource(resource_with_unknown_permissions, lookups)
        assert isinstance(result, ResourceInputConversionFailure)
        assert result.resource_id == "id"
        assert result.failure_msg == "Could not find permissions for value: nonExisting"


class TestTransformOneResource:
    def test_resource_one_prop(self, resource_one_prop: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_one_prop, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 1
        assert not result.file_value
        assert not result.migration_metadata

    def test_resource_with_permissions(self, resource_with_permissions: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_permissions, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert isinstance(result.permissions, Permissions)
        assert len(result.values) == 0
        assert not result.file_value
        assert not result.migration_metadata

    def test_with_ark(self, resource_with_ark: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_ark, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/5d5d1FKaUC2Wfl4zicggfg"
        assert metadata.creation_date == "1999-12-31T23:59:59.9999999+01:00"

    def test_with_iri(self, resource_with_iri: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_iri, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA"
        assert not metadata.creation_date

    def test_resource_with_ark_and_iri(self, resource_with_ark_and_iri: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_ark_and_iri, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/5d5d1FKaUC2Wfl4zicggfg"
        time_stamp = metadata.creation_date
        assert isinstance(time_stamp, DateTimeStamp)
        assert time_stamp == DateTimeStamp("1999-12-31T23:59:59.9999999+01:00")

    def test_unknown_permission(self, resource_with_unknown_permissions: XMLResource, lookups: IntermediaryLookups):
        with pytest.raises(PermissionNotExistsError):
            _transform_one_resource(resource_with_unknown_permissions, lookups)

    def test_bitstream(self, resource_with_bitstream: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_bitstream, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        file_val = result.file_value
        assert isinstance(file_val, IntermediaryFileValue)
        assert file_val.value == "file.jpg"
        assert not file_val.metadata.permissions
        assert not result.migration_metadata

    def test_iiif_uri(self, resource_with_iiif_uri: XMLResource, lookups: IntermediaryLookups):
        result = _transform_one_resource(resource_with_iiif_uri, lookups)
        assert result.res_id == "id"
        assert result.type_iri == f"{ONTO}ResourceType"
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        assert not result.migration_metadata


class TestTransformFileValue:
    def test_transform_file_value(self, bitstream: XMLBitstream, lookups: IntermediaryLookups):
        result = _transform_file_value(bitstream, lookups, "id", "lbl")
        assert result.value == "file.jpg"
        assert isinstance(result, IntermediaryFileValue)
        metadata = result.metadata
        assert not metadata.permissions
        assert metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert metadata.copyright_holder == "copy"
        assert metadata.authorships == ["author"]

    def test_transform_file_value_with_permissions(
        self, bitstream_with_permission: XMLBitstream, lookups: IntermediaryLookups
    ):
        result = _transform_file_value(bitstream_with_permission, lookups, "id", "lbl")
        assert isinstance(result, IntermediaryFileValue)
        assert result.value == "file.jpg"
        metadata = result.metadata
        assert isinstance(metadata.permissions, Permissions)
        assert metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert metadata.copyright_holder == "copy"
        assert metadata.authorships == ["author"]

    def test_transform_iiif_uri_value(self, iiif_uri: IIIFUriInfo, lookups: IntermediaryLookups):
        result = _transform_iiif_uri_value(iiif_uri, lookups)
        assert result.value == "https://this/is/a/uri.jpg"
        assert isinstance(result, IntermediaryIIIFUri)
        metadata = result.metadata
        assert not metadata.permissions
        assert metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert metadata.copyright_holder == "copy"
        assert metadata.authorships == ["author"]

    def test_transform_iiif_uri_value_with_permission(
        self, iiif_uri_with_permission: IIIFUriInfo, lookups: IntermediaryLookups
    ):
        result = _transform_iiif_uri_value(iiif_uri_with_permission, lookups)
        assert isinstance(result, IntermediaryIIIFUri)
        assert result.value == "https://this/is/a/uri.jpg"
        metadata = result.metadata
        assert isinstance(metadata.permissions, Permissions)
        assert metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert metadata.copyright_holder == "copy"
        assert metadata.authorships == ["author"]

    def test_get_metadata_soon_deprecated_without_metadata(self, lookups):
        file = IIIFUriInfo("uri", XMLFileMetadata(None, None, None))
        result = _transform_iiif_uri_value(file, lookups)
        assert not result.metadata.permissions
        assert not result.metadata.license_iri
        assert not result.metadata.copyright_holder
        assert not result.metadata.authorships

    def test_get_metadata_raises(self, lookups):
        metadata = XMLFileMetadata("unknown", "copy", "auth_id")
        msg = regex.escape(
            "The license 'unknown' used for an image or iiif-uri is unknown. "
            "See documentation for accepted pre-defined licenses."
        )
        with pytest.raises(InputError, match=msg):
            _get_metadata(metadata, lookups)

    def test_get_metadata_unknown_author(self, lookups):
        metadata = XMLFileMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "unknown")
        msg = regex.escape(
            "The authorship id 'unknown' referenced in a multimedia file or iiif-uri is unknown. "
            "Ensure that all referenced ids are defined in the `<authorship>` elements of your XML."
        )
        with pytest.raises(InputError, match=msg):
            _get_metadata(metadata, lookups)


class TestTransformProperties:
    def test_bool_prop(self, bool_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(bool_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryBoolean)
        assert transformed.value == True  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == f"{ONTO}boolProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_color_prop(self, color_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(color_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryColor)
        assert transformed.value == "#5d1f1e"
        assert transformed.prop_iri == f"{ONTO}colorProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_date_prop(self, date_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(date_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryDate)
        assert isinstance(transformed.value, Date)
        assert transformed.prop_iri == f"{ONTO}dateProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_decimal_prop(self, decimal_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(decimal_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryDecimal)
        assert transformed.value == 1.4
        assert transformed.prop_iri == f"{ONTO}decimalProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_decimal_prop_with_two_values(
        self, decimal_prop_with_two_values: XMLProperty, lookups: IntermediaryLookups
    ):
        result = _transform_one_property(decimal_prop_with_two_values, lookups)
        assert len(result) == 2

    def test_simple_text_prop(self, simple_text_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(simple_text_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediarySimpleText)
        assert transformed.value == "text"
        assert transformed.prop_iri == f"{ONTO}simpleTextProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_richtext_prop(self, richtext_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(richtext_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryRichtext)
        assert isinstance(transformed.value, FormattedTextValue)
        assert transformed.prop_iri == f"{ONTO}richTextProp"
        assert not transformed.permissions
        assert not transformed.comment
        assert transformed.resource_references == {"id"}

    def test_geoname_prop(self, geoname_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(geoname_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryGeoname)
        assert transformed.value == "5416656"
        assert transformed.prop_iri == f"{ONTO}geonameProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_integer_prop(self, integer_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(integer_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryInt)
        assert transformed.value == 1
        assert transformed.prop_iri == f"{ONTO}integerProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_list_prop(self, list_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(list_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryList)
        assert transformed.value == "http://rdfh.ch/9999/node"
        assert transformed.prop_iri == f"{ONTO}listProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_resptr_prop(self, resptr_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(resptr_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryLink)
        assert transformed.value == "other_id"
        assert transformed.prop_iri == f"{ONTO}linkPropValue"
        assert not transformed.permissions
        assert not transformed.comment

    def test_time_prop(self, time_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(time_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryTime)
        assert transformed.value == "2019-10-23T13:45:12.01-14:00"
        assert transformed.prop_iri == f"{ONTO}timeProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_uri_prop(self, uri_prop: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(uri_prop, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert isinstance(transformed, IntermediaryUri)
        assert transformed.value == "https://dasch.swiss"
        assert transformed.prop_iri == f"{ONTO}uriProp"
        assert not transformed.permissions
        assert not transformed.comment

    def test_bool_prop_with_comment(self, bool_prop_with_comment: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(bool_prop_with_comment, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert transformed.value == True  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == f"{ONTO}boolProp"
        assert not transformed.permissions
        assert transformed.comment == "comment"

    def test_bool_prop_with_permissions(self, bool_prop_with_permissions: XMLProperty, lookups: IntermediaryLookups):
        result = _transform_one_property(bool_prop_with_permissions, lookups)
        assert len(result) == 1
        transformed = result[0]
        assert transformed.value == True  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == f"{ONTO}boolProp"
        assert isinstance(transformed.permissions, Permissions)
        assert not transformed.comment

    def test_bool_prop_with_non_existing_permissions(
        self, bool_prop_with_non_existing_permissions: XMLProperty, lookups: IntermediaryLookups
    ):
        with pytest.raises(PermissionNotExistsError):
            _transform_one_property(bool_prop_with_non_existing_permissions, lookups)



# mypy: disable-error-code="method-assign,no-untyped-def"
import datetime

import pandas as pd
import pytest
import regex

from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.error.exceptions import InputError
from dsp_tools.xmllib.helpers import ListLookup
from dsp_tools.xmllib.helpers import create_footnote_string
from dsp_tools.xmllib.helpers import create_list_from_string
from dsp_tools.xmllib.helpers import create_non_empty_list_from_string
from dsp_tools.xmllib.helpers import create_standoff_link_to_resource
from dsp_tools.xmllib.helpers import create_standoff_link_to_uri
from dsp_tools.xmllib.helpers import escape_reserved_xml_characters
from dsp_tools.xmllib.helpers import find_date_in_string
from dsp_tools.xmllib.models.config_options import NewlineReplacement


@pytest.fixture
def list_lookup() -> ListLookup:
    return ListLookup(
        _lookup={
            "list1": {"Label 1": "list1_node1", "Label 2": "list1_node2"},
            "list2": {"Label 1": "list2_node1", "Label 2": "list2_node2"},
        },
        _prop_to_list_name={
            "default:defaultOntoHasListOne": "list1",
            ":defaultOntoHasListOne": "list1",
            "other-onto:otherOntoHasListOne": "list1",
            "other-onto:otherOntoHasListTwo": "list2",
        },
        _label_language="en",
    )


class TestFootnotes:
    @pytest.mark.parametrize(
        ("input_text", "newline_replacement", "expected"),
        [
            ("Text", NewlineReplacement.NONE, "Text"),
            ("With escape &", NewlineReplacement.NONE, "With escape &amp;"),
            ("With escape <", NewlineReplacement.NONE, "With escape &lt;"),
            ("With escape >", NewlineReplacement.NONE, "With escape &gt;"),
            ("<With escape>", NewlineReplacement.LINEBREAK, "&lt;With escape&gt;"),
            ("Already escaped &lt;&gt;", NewlineReplacement.LINEBREAK, "Already escaped &lt;&gt;"),
            (
                "Non-ASCII àéèêëôû äöüß 0123456789 _-'()+=!?| 漢が글ርبيةб中",
                NewlineReplacement.NONE,
                "Non-ASCII àéèêëôû äöüß 0123456789 _-'()+=!?| 漢が글ርبيةб中",
            ),
        ],
    )
    def test_create_footnote_string_correct(
        self, input_text: str, newline_replacement: NewlineReplacement, expected: str
    ) -> None:
        result = create_footnote_string(input_text, newline_replacement)
        expected = f'<footnote content="{expected}"/>'
        assert result == expected

    @pytest.mark.parametrize(
        ("input_text", "newline_replacement", "expected_msg"),
        [
            (
                "Text",
                NewlineReplacement.PARAGRAPH,
                "Currently the only supported newline replacement is linebreak (<br/>) or None.",
            ),
            (pd.NA, NewlineReplacement.NONE, "The input value is empty."),
        ],
    )
    def test_create_footnote_string_raises(
        self, input_text: str, newline_replacement: NewlineReplacement, expected_msg: str
    ) -> None:
        with pytest.raises(InputError, match=regex.escape(expected_msg)):
            create_footnote_string(input_text, newline_replacement)


def test_create_standoff_link_to_resource_success() -> None:
    result = create_standoff_link_to_resource("id", "Text")
    assert result == '<a class="salsah-link" href="IRI:id:IRI">Text</a>'


def test_create_standoff_link_to_resource_id_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_resource("", "Text")


def test_create_standoff_link_to_resource_text_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_resource("id", "")


def test_create_standoff_link_to_uri_success() -> None:
    result = create_standoff_link_to_uri("https://uri.ch", "Text")
    assert result == '<a href="https://uri.ch">Text</a>'


def test_create_standoff_link_to_uri_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_uri("", "Text")


def test_create_standoff_link_to_uri_text_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_uri("https://uri.ch", "")


class TestFindDate:
    def test_find_date_in_string_ignore_subsequent(self) -> None:
        assert find_date_in_string("x 1492-10-12, 2025-01-01x") == "GREGORIAN:CE:1492-10-12:CE:1492-10-12"
        assert find_date_in_string("first date: 2024. Second is ignored: 2025.") == "GREGORIAN:CE:2024:CE:2024"

    def test_find_date_in_string_accept_only_dash_as_range_delimiter(self) -> None:
        assert find_date_in_string("01.01.1900:31.12.2000") == "GREGORIAN:CE:1900-01-01:CE:1900-01-01"
        assert find_date_in_string("01.01.1900 to 31.12.2000") == "GREGORIAN:CE:1900-01-01:CE:1900-01-01"
        assert find_date_in_string("1900:2000") == "GREGORIAN:CE:1900:CE:1900"
        assert find_date_in_string("1900 to 2000") == "GREGORIAN:CE:1900:CE:1900"

    def test_find_date_in_string_iso(self) -> None:
        """template: 2021-01-01"""
        assert find_date_in_string("x 1492-10-12, x") == "GREGORIAN:CE:1492-10-12:CE:1492-10-12"
        assert find_date_in_string("x 0476-09-04. x") == "GREGORIAN:CE:0476-09-04:CE:0476-09-04"
        assert find_date_in_string("x (0476-09-04) x") == "GREGORIAN:CE:0476-09-04:CE:0476-09-04"
        assert find_date_in_string("x [1492-10-32?] x") is None

    def test_find_date_in_string_eur_date(self) -> None:
        """template: 31.4.2021 | 5/11/2021 | 2015_01_02"""
        assert find_date_in_string("x (30.4.2021) x") == "GREGORIAN:CE:2021-04-30:CE:2021-04-30"
        assert find_date_in_string("x (5/11/2021) x") == "GREGORIAN:CE:2021-11-05:CE:2021-11-05"
        assert find_date_in_string("x ...2193_01_26... x") == "GREGORIAN:CE:2193-01-26:CE:2193-01-26"
        assert find_date_in_string("x -2193_01_26- x") == "GREGORIAN:CE:2193-01-26:CE:2193-01-26"
        assert find_date_in_string("x 2193_02_30 x") is None

    def test_find_date_in_string_eur_date_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_date_in_string(f"x 30.4.{cur} x") == f"GREGORIAN:CE:20{cur}-04-30:CE:20{cur}-04-30"
        assert find_date_in_string(f"x 30.4.{nxt} x") == f"GREGORIAN:CE:19{nxt}-04-30:CE:19{nxt}-04-30"
        assert find_date_in_string(f"x 31.4.{nxt} x") is None

    def test_find_date_in_string_eur_date_range(self) -> None:
        """template: 27.-28.1.1900"""
        assert find_date_in_string("x 25.-26.2.0800 x") == "GREGORIAN:CE:0800-02-25:CE:0800-02-26"
        assert find_date_in_string("x 25. - 26.2.0800 x") == "GREGORIAN:CE:0800-02-25:CE:0800-02-26"
        assert find_date_in_string("x 25.-25.2.0800 x") == "GREGORIAN:CE:0800-02-25:CE:0800-02-25"
        assert find_date_in_string("x 25.-24.2.0800 x") is None

    def test_find_date_in_string_eur_date_range_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_date_in_string(f"x 15.-16.4.{cur} x") == f"GREGORIAN:CE:20{cur}-04-15:CE:20{cur}-04-16"
        assert find_date_in_string(f"x 15.-16.4.{nxt} x") == f"GREGORIAN:CE:19{nxt}-04-15:CE:19{nxt}-04-16"

    def test_find_date_in_string_eur_date_range_across_month(self) -> None:
        """template: 26.2.-24.3.1948"""
        assert find_date_in_string("x _1.3. - 25.4.2022_ x") == "GREGORIAN:CE:2022-03-01:CE:2022-04-25"
        assert find_date_in_string("x (01.03. - 25.04.2022) x") == "GREGORIAN:CE:2022-03-01:CE:2022-04-25"
        assert find_date_in_string("x 28.2.-1.12.1515 x") == "GREGORIAN:CE:1515-02-28:CE:1515-12-01"
        assert find_date_in_string("x 28.2.-28.2.1515 x") == "GREGORIAN:CE:1515-02-28:CE:1515-02-28"
        assert find_date_in_string("x 28.2.-26.2.1515 x") is None

    def test_find_date_in_string_eur_date_range_across_month_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_date_in_string(f"x 15.04.-1.5.{cur} x") == f"GREGORIAN:CE:20{cur}-04-15:CE:20{cur}-05-01"
        assert find_date_in_string(f"x 15.04.-1.5.{nxt} x") == f"GREGORIAN:CE:19{nxt}-04-15:CE:19{nxt}-05-01"

    def test_find_date_in_string_eur_date_range_across_year(self) -> None:
        """template: 1.12.1973 - 6.1.1974"""
        assert find_date_in_string("x 1.9.2022-3.1.2024 x") == "GREGORIAN:CE:2022-09-01:CE:2024-01-03"
        assert find_date_in_string("x 25.12.2022 - 3.1.2024 x") == "GREGORIAN:CE:2022-12-25:CE:2024-01-03"
        assert find_date_in_string("x 25/12/2022-03/01/2024 x") == "GREGORIAN:CE:2022-12-25:CE:2024-01-03"
        assert find_date_in_string("x 25/12/2022 - 3/1/2024 x") == "GREGORIAN:CE:2022-12-25:CE:2024-01-03"
        assert find_date_in_string("x 25.12.2022-25.12.2022 x") == "GREGORIAN:CE:2022-12-25:CE:2022-12-25"
        assert find_date_in_string("x 25/12/2022-25/12/2022 x") == "GREGORIAN:CE:2022-12-25:CE:2022-12-25"
        assert find_date_in_string("x 25.12.2022-03.01.2022 x") is None
        assert find_date_in_string("x 25/12/2022-03/01/2022 x") is None

    def test_find_date_in_string_eur_date_range_across_year_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_date_in_string(f"x 15.04.23-1.5.{cur} x") == f"GREGORIAN:CE:2023-04-15:CE:20{cur}-05-01"
        assert find_date_in_string(f"x 15.04.{nxt}-1.5.26 x") == f"GREGORIAN:CE:19{nxt}-04-15:CE:1926-05-01"

    def test_find_date_in_string_monthname(self) -> None:
        """template: February 9, 1908 | Dec 5,1908"""
        assert find_date_in_string("x Jan 26, 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x February26,2051 x") == "GREGORIAN:CE:2051-02-26:CE:2051-02-26"
        assert find_date_in_string("x Sept 1, 1000 x") == "GREGORIAN:CE:1000-09-01:CE:1000-09-01"
        assert find_date_in_string("x October 01, 1000 x") == "GREGORIAN:CE:1000-10-01:CE:1000-10-01"
        assert find_date_in_string("x Nov 6,1000 x") == "GREGORIAN:CE:1000-11-06:CE:1000-11-06"

    def test_find_date_in_string_monthname_after_day(self) -> None:
        """template: 15 Jan 1927 | 15 January 1927"""
        assert find_date_in_string("x 15 Jan 1927 x") == "GREGORIAN:CE:1927-01-15:CE:1927-01-15"
        assert find_date_in_string("x 15Jan1927 x") == "GREGORIAN:CE:1927-01-15:CE:1927-01-15"
        assert find_date_in_string("x 15 January 1927 x") == "GREGORIAN:CE:1927-01-15:CE:1927-01-15"
        assert find_date_in_string("x 15January1927 x") == "GREGORIAN:CE:1927-01-15:CE:1927-01-15"

    def test_find_date_in_string_german_monthnames(self) -> None:
        """template: 26. Jan. 1993 | 26. Januar 1993"""
        assert find_date_in_string("x 26 Jan 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26. Jan 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26 Jan. 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26. Jan. 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26.Jan. 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26.Jan.1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26Jan1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"
        assert find_date_in_string("x 26. Januar 1993 x") == "GREGORIAN:CE:1993-01-26:CE:1993-01-26"

    def test_find_date_in_string_single_year(self) -> None:
        """template: 1907 | 476"""
        assert find_date_in_string("Text 1848 text") == "GREGORIAN:CE:1848:CE:1848"
        assert find_date_in_string("Text 0476 text") == "GREGORIAN:CE:476:CE:476"
        assert find_date_in_string("Text 476 text") == "GREGORIAN:CE:476:CE:476"

    def test_find_date_in_string_year_range(self) -> None:
        """template: 1849/50 | 1845-50 | 1849/1850"""
        assert find_date_in_string("x 1849/1850? x") == "GREGORIAN:CE:1849:CE:1850"
        assert find_date_in_string("x 1845-1850, x") == "GREGORIAN:CE:1845:CE:1850"
        assert find_date_in_string("x 800-900, x") == "GREGORIAN:CE:800:CE:900"
        assert find_date_in_string("x 840-50, x") == "GREGORIAN:CE:840:CE:850"
        assert find_date_in_string("x 844-8, x") == "GREGORIAN:CE:844:CE:848"
        assert find_date_in_string("x 1840-1, x") == "GREGORIAN:CE:1840:CE:1841"
        assert find_date_in_string("x 0750-0760 x") == "GREGORIAN:CE:750:CE:760"
        assert find_date_in_string("x 1849/50. x") == "GREGORIAN:CE:1849:CE:1850"
        assert find_date_in_string("x (1845-50) x") == "GREGORIAN:CE:1845:CE:1850"
        assert find_date_in_string("x [1849/1850] x") == "GREGORIAN:CE:1849:CE:1850"
        assert find_date_in_string("x 1850-1849 x") is None
        assert find_date_in_string("x 1850-1850 x") is None
        assert find_date_in_string("x 830-20 x") is None
        assert find_date_in_string("x 830-30 x") is None
        assert find_date_in_string("x 1811-10 x") is None
        assert find_date_in_string("x 1811-11 x") is None
        assert find_date_in_string("x 1811/10 x") is None
        assert find_date_in_string("x 1811/11 x") is None

    @pytest.mark.parametrize("string", ["9 BC", "9 B.C.", "9 BCE", "9 B.C.E."])
    def test_find_date_in_string_bc_different_notations(self, string: str) -> None:
        assert find_date_in_string(string) == "GREGORIAN:BC:9:BC:9"

    @pytest.mark.parametrize("string", ["9 CE", "9 C.E.", "9 AD", "9 A.D."])
    def test_find_date_in_string_ce_different_notations(self, string: str) -> None:
        assert find_date_in_string(string) == "GREGORIAN:CE:9:CE:9"

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("x 9 BC x", "GREGORIAN:BC:9:BC:9"),
            ("x 10000 BC x", "GREGORIAN:BC:10000:BC:10000"),
            ("x 170 BC - 90 BC x", "GREGORIAN:BC:170:BC:90"),
            ("x 170-90 BCE x", "GREGORIAN:BC:170:BC:90"),
            ("x 20 BCE-50 CE x", "GREGORIAN:BC:20:CE:50"),
            ("x 20 BCE - 50 C.E. x", "GREGORIAN:BC:20:CE:50"),
        ],
    )
    def test_find_date_in_string_bc(self, string: str, expected: str) -> None:
        assert find_date_in_string(string) == expected

    def test_find_date_in_string_french_bc(self) -> None:
        assert find_date_in_string("Text 12345 av. J.-C. text") == "GREGORIAN:BC:12345:BC:12345"
        assert find_date_in_string("Text 2000 av. J.-C. text") == "GREGORIAN:BC:2000:BC:2000"
        assert find_date_in_string("Text 250 av. J.-C. text") == "GREGORIAN:BC:250:BC:250"
        assert find_date_in_string("Text 33 av. J.-C. text") == "GREGORIAN:BC:33:BC:33"
        assert find_date_in_string("Text 1 av. J.-C. text") == "GREGORIAN:BC:1:BC:1"

    def test_find_date_in_string_french_bc_ranges(self) -> None:
        assert find_date_in_string("Text 99999-1000 av. J.-C. text") == "GREGORIAN:BC:99999:BC:1000"
        assert find_date_in_string("Text 1125-1050 av. J.-C. text") == "GREGORIAN:BC:1125:BC:1050"
        assert find_date_in_string("Text 1234-987 av. J.-C. text") == "GREGORIAN:BC:1234:BC:987"
        assert find_date_in_string("Text 350-340 av. J.-C. text") == "GREGORIAN:BC:350:BC:340"
        assert find_date_in_string("Text 842-98 av. J.-C. text") == "GREGORIAN:BC:842:BC:98"
        assert find_date_in_string("Text 45-26 av. J.-C. text") == "GREGORIAN:BC:45:BC:26"
        assert find_date_in_string("Text 53-7 av. J.-C. text") == "GREGORIAN:BC:53:BC:7"
        assert find_date_in_string("Text 6-5 av. J.-C. text") == "GREGORIAN:BC:6:BC:5"

    def test_find_date_in_string_french_bc_orthographical_variants(self) -> None:
        assert find_date_in_string("Text 1 av. J.-C. text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av J.-C. text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av.J.-C. text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av. J.C. text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av. J-C text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av.JC text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av JC text") == "GREGORIAN:BC:1:BC:1"
        assert find_date_in_string("Text 1 av. J.-C.text") == "GREGORIAN:BC:1:BC:1"

    def test_find_date_in_string_french_bc_dash_variants(self) -> None:
        assert find_date_in_string("Text 2000-1000 av. J.-C. text") == "GREGORIAN:BC:2000:BC:1000"
        assert find_date_in_string("Text 2000- 1000 av. J.-C. text") == "GREGORIAN:BC:2000:BC:1000"
        assert find_date_in_string("Text 2000 -1000 av. J.-C. text") == "GREGORIAN:BC:2000:BC:1000"
        assert find_date_in_string("Text 2000 - 1000 av. J.-C. text") == "GREGORIAN:BC:2000:BC:1000"

    def test_find_date_in_string_french_bc_invalid_syntax(self) -> None:
        assert find_date_in_string("Text12 av. J.-C. text") is None
        assert find_date_in_string("Text 12 av. J.-Ctext") is None
        assert find_date_in_string("Text 1 avJC text") is None

    def test_find_date_in_string_french_bc_invalid_range(self) -> None:
        assert find_date_in_string("Text 12-20 av. J.-C. text") is None


class TestCreateListFromString:
    def test_create_list_from_string_ok(self) -> None:
        result = create_list_from_string("ab, cd , ", ",")
        assert set(result) == {"ab", "cd"}

    def test_create_list_from_string_not_string(self) -> None:
        msg = regex.escape("The input for this function must be a string. Your input is a bool.")
        with pytest.raises(InputError, match=msg):
            create_list_from_string(True, ",")  # type: ignore[arg-type]

    def test_create_list_from_string_empty(self) -> None:
        result = create_list_from_string(" , ", ",")
        assert isinstance(result, list)
        assert result == []

    def test_create_non_empty_list_from_string_ok(self) -> None:
        result = create_non_empty_list_from_string("ab, cd , ", ",")
        assert set(result) == {"ab", "cd"}

    def test_create_non_empty_list_from_string_raises(self) -> None:
        msg = regex.escape(
            "The input for this function must result in a non-empty list. Your input results in an empty list."
        )
        with pytest.raises(InputError, match=msg):
            create_non_empty_list_from_string(" , ", ",")


class TestListLookup:
    def test_get_node_via_list_name(self, list_lookup):
        assert list_lookup.get_node_via_list_name("list1", "Label 1") == "list1_node1"

    def test_get_node_via_list_name_warns_wrong_list(self, list_lookup):
        msg = regex.escape("Entered list name 'inexistent' was not found.")
        with pytest.warns(DspToolsUserWarning, match=msg):
            result = list_lookup.get_node_via_list_name("inexistent", "Label 1")
        assert result == ""

    def test_get_node_via_list_name_warns_wrong_node(self, list_lookup):
        msg = regex.escape(
            "'inexistent' was not recognised as label of the list 'list1'. "
            "This ListLookup is configured for 'en' labels."
        )
        with pytest.warns(DspToolsUserWarning, match=msg):
            result = list_lookup.get_node_via_list_name("list1", "inexistent")
        assert result == ""

    def test_get_node_via_property(self, list_lookup):
        list_name, node_name = list_lookup.get_list_name_and_node_via_property(
            "other-onto:otherOntoHasListTwo", "Label 2"
        )
        assert list_name == "list2"
        assert node_name == "list2_node2"

    def test_get_node_via_property_warns_wrong_property(self, list_lookup):
        msg = regex.escape("Entered property ':inexistent' was not found.")
        with pytest.warns(DspToolsUserWarning, match=msg):
            result = list_lookup.get_list_name_and_node_via_property(":inexistent", "Label 2")
        assert result == ("", "")


@pytest.mark.parametrize(
    ("input_val", "expected"),
    [
        ("Text no escape", "Text no escape"),
        ("known tag <strong>content</strong>", "known tag <strong>content</strong>"),
        ("Ampersand &", "Ampersand &amp;"),
        ("Unknow tags <unknown></unknown>", "Unknow tags &lt;unknown&gt;&lt;/unknown&gt;"),
        ("<text in brackets>", "&lt;text in brackets&gt;"),
    ],
)
def test_escape_reserved_xml_characters(input_val: str, expected: str) -> None:
    result = escape_reserved_xml_characters(input_val)
    assert result == expected



import numpy as np
import pandas as pd
import pytest

from dsp_tools.xmllib.internal_helpers import is_nonempty_value_internal


@pytest.mark.parametrize(
    ("text", "expected"),
    [
        (np.nan, False),
        (pd.NA, False),
        (None, False),
        ("", False),
        (" ", False),
        (" \t ", False),
        (" \n ", False),
        (" \r ", False),
        (" \v ", False),  # vertical tab
        (" \f ", False),
        (" \u00a0 ", False),  # non-breaking space
        (" \u200b ", False),  # zero-width space
        (" \ufeff ", False),  # Zero-Width No-Break Space
        (" \t\n\r\v\f \u00a0 \u200b \ufeff", False),
        ("a", True),
        (0, True),
        (1, True),
        ("0", True),
        ("1", True),
        (True, True),
        (False, True),
    ],
)
def test_is_nonempty_value_internal(text: str, expected: bool) -> None:
    assert is_nonempty_value_internal(text) == expected



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import SHACLListInfo
from dsp_tools.commands.validate_data.sparql.value_shacl import _add_property_shapes_to_class_shapes
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_link_value_shape
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_link_value_type_shapes_to_class_shapes
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_one_list_node_shape
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_one_list_property_shape_with_collection
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_one_property_type_text_value
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_value_type_shapes_to_class_shapes
from dsp_tools.utils.rdflib_constants import API_SHAPES
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.unittests.commands.validate_data.constants import ONTO


def test_construct_link_value_shape(link_prop: Graph) -> None:
    res = _construct_link_value_shape(link_prop)
    assert len(res) == 3
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, SH.path)) == ONTO.testHasLinkTo
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, RDF.type)) == SH.PropertyShape
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, SH.node)) == ONTO.testHasLinkTo_NodeShape


def test_construct_one_property_type_text_value(one_richtext_prop: Graph) -> None:
    res = _construct_one_property_type_text_value(
        one_richtext_prop, "salsah-gui:Richtext", "api-shapes:FormattedTextValue_ClassShape"
    )
    assert len(res) == 3
    assert next(res.objects(ONTO.testRichtext_PropShape, SH.path)) == ONTO.testRichtext
    assert next(res.objects(ONTO.testRichtext_PropShape, RDF.type)) == SH.PropertyShape
    assert next(res.objects(ONTO.testRichtext_PropShape, SH.node)) == API_SHAPES.FormattedTextValue_ClassShape


def test_add_property_shapes_to_class_shapes(card_1: Graph) -> None:
    res = _add_property_shapes_to_class_shapes(card_1)
    expected_props = {ONTO.testBoolean_PropShape}
    assert set(res.objects(ONTO.ClassMixedCard, SH.property)) == expected_props


def test_construct_value_type_shapes_to_class_shapes_literal(card_1: Graph) -> None:
    res = _construct_value_type_shapes_to_class_shapes(card_1)
    prop_iri = ONTO.testBoolean_PropShape
    assert next(res.objects(prop_iri, SH.path)) == ONTO.testBoolean
    assert next(res.objects(prop_iri, URIRef("http://www.w3.org/ns/shacl#class"))) == KNORA_API.BooleanValue
    assert next(res.objects(prop_iri, SH.message)) == Literal("This property requires a BooleanValue")
    assert len(res) == 4


def test_construct_value_type_shapes_to_class_shapes_link_value(link_prop_card_1: Graph) -> None:
    res = _construct_value_type_shapes_to_class_shapes(link_prop_card_1)
    assert len(res) == 0


def test_construct_link_value_type_shapes_to_class_shapes_literal(card_1: Graph) -> None:
    res = _construct_link_value_type_shapes_to_class_shapes(card_1)
    assert len(res) == 0


def test_construct_link_value_type_shapes_to_class_shapes_link_value(link_prop_card_1: Graph) -> None:
    res = _construct_link_value_type_shapes_to_class_shapes(link_prop_card_1)
    prop_iri = ONTO.testHasLinkToCardOneResource_PropShape
    assert next(res.objects(prop_iri, SH.path)) == ONTO.testHasLinkToCardOneResource
    assert next(res.objects(prop_iri, URIRef("http://www.w3.org/ns/shacl#class"))) == KNORA_API.LinkValue
    assert next(res.objects(prop_iri, SH.message)) == Literal("This property requires a LinkValue")
    assert len(res) == 4


class TestConstructListNode:
    def test_node_space(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name="list",
            nodes=["l2n1 space"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_node_backslash(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name="list",
            nodes=["l2n1 \\ or"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_node_double_quote(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name="list",
            nodes=['l2n2"'],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_node_apostrophe(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name="list",
            nodes=["l2n3'"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_list_special(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name="secondList \\ ",
            nodes=["a"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_list_double_quote(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name='secondList " ',
            nodes=["a"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_list_single_quote(self) -> None:
        test_list = OneList(
            list_iri="http://rdfh.ch/lists/9999/test",
            list_name='secondList " ',
            nodes=["a"],
        )
        result = _construct_one_list_node_shape(test_list)
        nodeshape_iri = URIRef("http://rdfh.ch/lists/9999/test")
        assert next(result.subjects(RDF.type, SH.NodeShape)) == nodeshape_iri

    def test_construct_one_list_property_shape_with_collection_one(self) -> None:
        test_info = SHACLListInfo(
            list_iri=URIRef("http://rdfh.ch/lists/9999/test"),
            sh_path=ONTO.testListProp,
            sh_message="Test",
            sh_in=["one"],
        )
        result = _construct_one_list_property_shape_with_collection(test_info)
        number_of_strings_in_list = 1
        assert len(list(result.objects(predicate=RDF.first))) == number_of_strings_in_list

    def test_construct_one_list_property_shape_with_collection_three(self) -> None:
        test_info = SHACLListInfo(
            list_iri=URIRef("http://rdfh.ch/lists/9999/test"),
            sh_path=ONTO.testListProp,
            sh_message="Test",
            sh_in=["one", "two", "three"],
        )
        result = _construct_one_list_property_shape_with_collection(test_info)
        number_of_strings_in_list = 3
        assert len(list(result.objects(predicate=RDF.first))) == number_of_strings_in_list


if __name__ == "__main__":
    pytest.main([__file__])



import json
import re
import urllib
from pathlib import Path

import pytest
from requests import RequestException
from requests_mock import Mocker

from dsp_tools.commands.ingest_xmlupload.bulk_ingest_client import BulkIngestClient
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import InputError
from test.integration.commands.xmlupload.authentication_client_mock import AuthenticationClientMockBase

DSP_INGEST_URL = "https://example.com"
SHORTCODE = "0001"


@pytest.fixture
def ingest_client() -> BulkIngestClient:
    return BulkIngestClient(DSP_INGEST_URL, AuthenticationClientMockBase(), SHORTCODE)


@pytest.fixture
def tmp_file(tmp_path: Path) -> Path:
    return tmp_path / "filename_îïôœùûüÿ.xml"


def _make_url(file: Path) -> str:
    filename = str(file)[1:] if str(file).startswith("/") else str(file)
    filename = urllib.parse.quote(filename, safe="")
    return f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest/ingest/{filename}"


def test_upload_file_success(ingest_client: BulkIngestClient, requests_mock: Mocker, tmp_file: Path) -> None:
    file_content = "<xml></xml>"
    tmp_file.write_text(file_content)
    url = _make_url(tmp_file)
    requests_mock.post(url, status_code=200)
    failure_detail = ingest_client.upload_file(tmp_file)
    assert not failure_detail
    assert len(requests_mock.request_history) == 1
    req = requests_mock.request_history[0]
    assert req.url == url
    assert req.method == "POST"
    assert req.headers["Authorization"] == "Bearer mocked_token"
    assert req.headers["Content-Type"] == "application/octet-stream"


def test_upload_file_with_inexisting_file(ingest_client: BulkIngestClient) -> None:
    failure_detail = ingest_client.upload_file(Path("inexisting.xml"))
    assert failure_detail
    assert failure_detail.filepath == Path("inexisting.xml")
    assert re.search(r"the file could not be opened/read", failure_detail.reason)


def test_upload_file_failure_upon_request_exception(
    ingest_client: BulkIngestClient, requests_mock: Mocker, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(tmp_file), exc=RequestException("Test exception"))
    failure_detail = ingest_client.upload_file(tmp_file)
    assert failure_detail
    assert failure_detail.filepath == tmp_file
    assert failure_detail.reason == "Exception of requests library: Test exception"


def test_upload_file_failure_upon_server_error(
    ingest_client: BulkIngestClient, requests_mock: Mocker, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(tmp_file), status_code=500)
    failure_detail = ingest_client.upload_file(tmp_file)
    assert failure_detail
    assert failure_detail.filepath == tmp_file
    assert failure_detail.status_code == 500


def test_upload_file_failure_upon_server_error_with_response_text(
    ingest_client: BulkIngestClient, requests_mock: Mocker, tmp_file: Path
) -> None:
    tmp_file.write_text("<xml></xml>")
    requests_mock.post(_make_url(tmp_file), status_code=500, text="response text")
    failure_detail = ingest_client.upload_file(tmp_file)
    assert failure_detail
    assert failure_detail.filepath == tmp_file
    assert failure_detail.status_code == 500
    assert failure_detail.response_text == "response text"


@pytest.mark.parametrize(
    ("filepath", "url_suffix"),
    [
        (Path("Côté gauche/Süd.png"), "C%C3%B4t%C3%A9%20gauche%2FS%C3%BCd.png"),
        (Path("/absolute/path/to/file.txt"), "absolute%2Fpath%2Fto%2Ffile.txt"),
    ],
)
def test_build_url_for_bulk_ingest_ingest_route(
    ingest_client: BulkIngestClient, filepath: Path, url_suffix: str
) -> None:
    res = ingest_client._build_url_for_bulk_ingest_ingest_route(filepath)
    common_part = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest/ingest/"
    assert res == f"{common_part}{url_suffix}"


def test_trigger_if_success(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=202, text=json.dumps({"id": SHORTCODE}))
    ingest_client.trigger_ingest_process()
    assert len(requests_mock.request_history) == 1
    req = requests_mock.request_history[0]
    assert req.url == url
    assert req.method == "POST"
    assert req.headers["Authorization"] == "Bearer mocked_token"


def test_trigger_if_unauthorized(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=401)
    with pytest.raises(BadCredentialsError):
        ingest_client.trigger_ingest_process()


def test_trigger_if_forbidden(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=403)
    with pytest.raises(BadCredentialsError):
        ingest_client.trigger_ingest_process()


def test_trigger_wrong_shortcode(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    ingest_client.shortcode = "9999"
    url = f"{DSP_INGEST_URL}/projects/9999/bulk-ingest"
    requests_mock.post(url, status_code=404)
    err_msg = re.escape(
        "No assets have been uploaded for project 9999. "
        "Before using the 'ingest-files' command, you must upload some files with the 'upload-files' command."
    )
    with pytest.raises(InputError, match=err_msg):
        ingest_client.trigger_ingest_process()


def test_trigger_when_ingest_already_running(
    ingest_client: BulkIngestClient, requests_mock: Mocker, capsys: pytest.CaptureFixture[str]
) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=409)
    ingest_client.trigger_ingest_process()
    expected = f"Ingest process on the server {DSP_INGEST_URL} is already running. Wait until it completes..."
    captured = capsys.readouterr().out.removesuffix("\n")
    assert captured == expected


def test_trigger_on_internal_server_error(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=500)
    with pytest.raises(InputError, match=re.escape("Server is unavailable. Please try again later.")):
        ingest_client.trigger_ingest_process()


def test_trigger_on_server_unavailable(ingest_client: BulkIngestClient, requests_mock: Mocker) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=503)
    with pytest.raises(InputError, match=re.escape("Server is unavailable. Please try again later.")):
        ingest_client.trigger_ingest_process()


def test_trigger_when_server_response_doesnt_contain_right_shortcode(
    ingest_client: BulkIngestClient, requests_mock: Mocker
) -> None:
    url = f"{DSP_INGEST_URL}/projects/{SHORTCODE}/bulk-ingest"
    requests_mock.post(url, status_code=202)
    err_msg = re.escape("Failed to trigger the ingest process. Please check the server logs, or try again later.")
    with pytest.raises(InputError, match=err_msg):
        ingest_client.trigger_ingest_process()



from typing import cast

from rdflib import RDF
from rdflib import RDFS
from rdflib import XSD
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.xmlupload.make_rdf_graph.make_file_value import make_file_value_graph
from dsp_tools.commands.xmlupload.make_rdf_graph.make_file_value import make_iiif_uri_value_graph
from dsp_tools.commands.xmlupload.make_rdf_graph.make_values import make_values
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import BitstreamInfo
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileMetadata
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileValue
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.res import MigrationMetadata
from dsp_tools.commands.xmlupload.models.lookup_models import IRILookups
from dsp_tools.commands.xmlupload.models.rdf_models import AbstractFileValue
from dsp_tools.commands.xmlupload.models.rdf_models import FileValueMetadata
from dsp_tools.utils.rdflib_constants import KNORA_API


def create_resource_with_values(
    resource: IntermediaryResource,
    bitstream_information: BitstreamInfo | None,
    lookups: IRILookups,
) -> Graph:
    """
    This function takes an XMLResource and serialises it into a json-ld type dict that can be sent to the API.

    Args:
        resource: XMLResource
        bitstream_information: if the resource has a FileValue
        lookups: Lookups to resolve IRIs, etc.

    Returns:
        A resource as a graph
    """

    graph = Graph()

    res_node: BNode | URIRef = BNode()
    if migration := resource.migration_metadata:
        if migration.iri_str:
            res_node = URIRef(migration.iri_str)
        graph += _make_migration_metadata(migration, res_node)

    graph += _make_resource(resource=resource, res_node=res_node, project_iri=lookups.project_iri)

    graph += _make_values_graph_from_resource(
        resource=resource, res_node=res_node, bitstream_information=bitstream_information, lookups=lookups
    )

    return graph


def _make_values_graph_from_resource(
    resource: IntermediaryResource,
    res_node: BNode | URIRef,
    bitstream_information: BitstreamInfo | None,
    lookups: IRILookups,
) -> Graph:
    properties_graph = make_values(resource.values, res_node, lookups)

    if resource.iiif_uri:
        metadata = _make_file_value_metadata(resource.iiif_uri.metadata)
        iiif_g = make_iiif_uri_value_graph(AbstractFileValue(resource.iiif_uri.value, metadata), res_node)
        properties_graph += iiif_g

    elif bitstream_information:
        file_val = cast(IntermediaryFileValue, resource.file_value)
        metadata = _make_file_value_metadata(file_val.metadata)
        file_g = make_file_value_graph(bitstream_information, metadata, res_node)
        properties_graph += file_g

    return properties_graph


def _make_file_value_metadata(intermediary_metadata: IntermediaryFileMetadata) -> FileValueMetadata:
    permissions = None
    if found := intermediary_metadata.permissions:
        permissions = str(found)
    return FileValueMetadata(
        intermediary_metadata.license_iri,
        intermediary_metadata.copyright_holder,
        intermediary_metadata.authorships,
        permissions,
    )


def _make_resource(resource: IntermediaryResource, res_node: BNode | URIRef, project_iri: URIRef) -> Graph:
    g = Graph()
    g.add((res_node, RDF.type, URIRef(resource.type_iri)))
    g.add((res_node, RDFS.label, Literal(resource.label, datatype=XSD.string)))
    g.add((res_node, KNORA_API.attachedToProject, project_iri))
    if resource.permissions:
        g.add((res_node, KNORA_API.hasPermissions, Literal(str(resource.permissions), datatype=XSD.string)))
    return g


def _make_migration_metadata(migration_metadata: MigrationMetadata, res_node: BNode | URIRef) -> Graph:
    g = Graph()
    if date := migration_metadata.creation_date:
        g.add((res_node, KNORA_API.creationDate, Literal(str(date), datatype=XSD.dateTimeStamp)))
    return g



# mypy: disable-error-code="method-assign,no-untyped-def"
from copy import deepcopy

import pytest
from lxml import etree

from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.get_parsed_resources import _create_from_local_name_to_absolute_iri_lookup
from dsp_tools.utils.xml_parsing.get_parsed_resources import _get_file_value_type
from dsp_tools.utils.xml_parsing.get_parsed_resources import _get_one_absolute_iri
from dsp_tools.utils.xml_parsing.get_parsed_resources import _parse_file_values
from dsp_tools.utils.xml_parsing.get_parsed_resources import _parse_iiif_uri
from dsp_tools.utils.xml_parsing.get_parsed_resources import _parse_one_value
from dsp_tools.utils.xml_parsing.get_parsed_resources import _parse_segment_values
from dsp_tools.utils.xml_parsing.get_parsed_resources import get_parsed_resources
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedMigrationMetadata

API_URL = "http://url.ch"
DEFAULT_ONTO_NAMESPACE = f"{API_URL}/ontology/0000/default/v2#"

HAS_PROP = f"{DEFAULT_ONTO_NAMESPACE}hasProp"
RES_CLASS = f"{DEFAULT_ONTO_NAMESPACE}Class"

IRI_LOOKUP = {
    ":Class": RES_CLASS,
    ":hasProp": HAS_PROP,
    "hasColor": f"{KNORA_API_STR}hasColor",
    "isRegionOf": f"{KNORA_API_STR}isRegionOf",
    "hasGeometry": f"{KNORA_API_STR}hasGeometry",
    "hasComment": f"{KNORA_API_STR}hasComment",
    "hasLinkTo": f"{KNORA_API_STR}hasLinkTo",
}


class TestParseResource:
    def test_empty(self, root_no_resources, resource_no_values):
        root = deepcopy(root_no_resources)
        root.append(resource_no_values)
        parsed_res, iris = get_parsed_resources(root, API_URL)
        assert iris == {RES_CLASS}
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_no_values"
        assert resource.res_type == RES_CLASS
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 0
        assert not resource.file_value
        assert not resource.migration_metadata

    def test_resource_with_migration_data(self, root_no_resources, resource_with_migration_data):
        root = deepcopy(root_no_resources)
        root.append(resource_with_migration_data)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_with_migration_data"
        assert resource.res_type == RES_CLASS
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 0
        assert not resource.file_value
        metadata = resource.migration_metadata
        assert isinstance(metadata, ParsedMigrationMetadata)
        assert metadata.iri == "iri"
        assert metadata.ark == "ark"
        assert metadata.creation_date == "2019-01-09T15:45:54.502951Z"

    def test_with_value(self, root_no_resources, resource_with_value):
        root = deepcopy(root_no_resources)
        root.append(resource_with_value)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_with_value"
        assert resource.res_type == RES_CLASS
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 1
        val = resource.values.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "true"
        assert val.value_type == KnoraValueType.BOOLEAN_VALUE
        assert not val.permissions_id
        assert not val.comment
        assert not resource.file_value
        assert not resource.migration_metadata

    def test_file_value(self, root_no_resources, resource_with_file_value):
        root = deepcopy(root_no_resources)
        root.append(resource_with_file_value)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_with_file_value"
        assert resource.res_type == RES_CLASS
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 0
        assert not resource.migration_metadata
        file_val = resource.file_value
        assert isinstance(file_val, ParsedFileValue)
        assert file_val.value_type == KnoraValueType.AUDIO_FILE
        assert file_val.value == "testdata/bitstreams/test.wav"

    def test_iiif_value(self, root_no_resources, resource_with_iiif):
        root = deepcopy(root_no_resources)
        root.append(resource_with_iiif)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_with_iiif"
        assert resource.res_type == RES_CLASS
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 0
        assert not resource.migration_metadata
        file_val = resource.file_value
        assert isinstance(file_val, ParsedFileValue)
        assert file_val.value_type == KnoraValueType.STILL_IMAGE_IIIF

    def test_region(self, root_no_resources, resource_region):
        root = deepcopy(root_no_resources)
        root.append(resource_region)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_region"
        assert resource.res_type == f"{KNORA_API_STR}Region"
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 4
        assert not resource.file_value
        assert not resource.migration_metadata

    def test_link(self, root_no_resources, resource_link):
        root = deepcopy(root_no_resources)
        root.append(resource_link)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_link"
        assert resource.res_type == f"{KNORA_API_STR}LinkObj"
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 1
        assert not resource.file_value
        assert not resource.migration_metadata


class TestSegment:
    def test_video_segment(self, root_no_resources, resource_video_segment):
        root = deepcopy(root_no_resources)
        root.append(resource_video_segment)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_video_segment"
        assert resource.res_type == f"{KNORA_API_STR}VideoSegment"
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 8
        assert not resource.file_value
        assert not resource.migration_metadata

    def test_audio_segment(self, root_no_resources, resource_audio_segment):
        root = deepcopy(root_no_resources)
        root.append(resource_audio_segment)
        parsed_res, _ = get_parsed_resources(root, API_URL)
        assert len(parsed_res) == 1
        resource = parsed_res.pop(0)
        assert resource.res_id == "resource_audio_segment"
        assert resource.res_type == f"{KNORA_API_STR}AudioSegment"
        assert resource.label == "lbl"
        assert not resource.permissions_id
        assert len(resource.values) == 8
        assert not resource.file_value
        assert not resource.migration_metadata

    def test_parse_segment_values(self, resource_audio_segment):
        values = _parse_segment_values(resource_audio_segment, "Audio")
        expected = [
            (f"{KNORA_API_STR}isAudioSegmentOf", KnoraValueType.LINK_VALUE, "target", None, None),
            (f"{KNORA_API_STR}hasSegmentBounds", KnoraValueType.INTERVAL_VALUE, ("0.1", "0.234"), "open", None),
            (f"{KNORA_API_STR}hasTitle", KnoraValueType.SIMPLETEXT_VALUE, "Title", None, "Cmt"),
            (f"{KNORA_API_STR}hasComment", KnoraValueType.RICHTEXT_VALUE, "Comment", None, None),
            (f"{KNORA_API_STR}hasDescription", KnoraValueType.RICHTEXT_VALUE, "Description 1", None, None),
            (f"{KNORA_API_STR}hasDescription", KnoraValueType.RICHTEXT_VALUE, "Description 2", None, None),
            (f"{KNORA_API_STR}hasKeyword", KnoraValueType.SIMPLETEXT_VALUE, "Keyword", None, None),
            (f"{KNORA_API_STR}relatesTo", KnoraValueType.LINK_VALUE, "relates_to_id", None, None),
        ]
        for result, (prop, value_type, content, perm, cmt) in zip(values, expected):
            assert result.prop_name == prop
            assert result.value == content
            assert result.value_type == value_type
            assert result.permissions_id == perm
            assert result.comment == cmt


class TestParseValues:
    def test_boolean_value_with_metadata(self):
        xml_val = etree.fromstring("""
        <boolean-prop name=":hasProp">
            <boolean permissions="open" comment="Comment on Value">true</boolean>
        </boolean-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "true"
        assert val.value_type == KnoraValueType.BOOLEAN_VALUE
        assert val.permissions_id == "open"
        assert val.comment == "Comment on Value"

    def test_color_value(self):
        xml_val = etree.fromstring("""
        <color-prop name=":hasProp">
            <color>#00ff00</color>
        </color-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "#00ff00"
        assert val.value_type == KnoraValueType.COLOR_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_color_value_several(self):
        xml_val = etree.fromstring("""
        <color-prop name=":hasProp">
            <color>#00ff00</color>
            <color>#00ff11</color>
        </color-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 2
        val1 = result[0]
        assert val1.prop_name == HAS_PROP
        assert val1.value == "#00ff00"
        assert val1.value_type == KnoraValueType.COLOR_VALUE
        assert not val1.permissions_id
        assert not val1.comment
        val2 = result[1]
        assert val2.prop_name == HAS_PROP
        assert val2.value == "#00ff11"
        assert val2.value_type == KnoraValueType.COLOR_VALUE
        assert not val2.permissions_id
        assert not val2.comment

    def test_date_value(self):
        xml_val = etree.fromstring("""
        <date-prop name=":hasProp">
            <date>JULIAN:BCE:0700:BCE:0600</date>
        </date-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "JULIAN:BCE:0700:BCE:0600"
        assert val.value_type == KnoraValueType.DATE_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_decimal_value(self):
        xml_val = etree.fromstring("""
        <decimal-prop name=":hasProp">
            <decimal>2.71</decimal>
        </decimal-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "2.71"
        assert val.value_type == KnoraValueType.DECIMAL_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_geometry_value(self):
        xml_val = etree.fromstring("""
        <geometry-prop name="hasGeometry">
            <geometry>{}</geometry>
        </geometry-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == f"{KNORA_API_STR}hasGeometry"
        assert val.value == "{}"
        assert val.value_type == KnoraValueType.GEOM_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_geoname_value(self):
        xml_val = etree.fromstring("""
        <geoname-prop name=":hasProp">
            <geoname>1111111</geoname>
        </geoname-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "1111111"
        assert val.value_type == KnoraValueType.GEONAME_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_integer_value(self):
        xml_val = etree.fromstring("""
        <integer-prop name=":hasProp">
            <integer>1</integer>
        </integer-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "1"
        assert val.value_type == KnoraValueType.INT_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_list_value(self):
        xml_val = etree.fromstring("""
        <list-prop list="firstList" name=":hasProp">
            <list>n1</list>
        </list-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == ("firstList", "n1")
        assert val.value_type == KnoraValueType.LIST_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_list_value_several(self):
        xml_val = etree.fromstring("""
        <list-prop list="firstList" name=":hasProp">
            <list>n1</list>
            <list>n2</list>
        </list-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 2
        val1 = result[0]
        assert val1.prop_name == HAS_PROP
        assert val1.value == ("firstList", "n1")
        assert val1.value_type == KnoraValueType.LIST_VALUE
        assert not val1.permissions_id
        assert not val1.comment
        val2 = result[1]
        assert val2.prop_name == HAS_PROP
        assert val2.value == ("firstList", "n2")
        assert val2.value_type == KnoraValueType.LIST_VALUE
        assert not val2.permissions_id
        assert not val2.comment

    def test_resptr_value(self):
        xml_val = etree.fromstring("""
        <resptr-prop name=":hasProp">
            <resptr>id_1</resptr>
        </resptr-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "id_1"
        assert val.value_type == KnoraValueType.LINK_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_text_richtext_value(self):
        xml_val = etree.fromstring("""
        <text-prop name=":hasProp">
            <text encoding="xml"><p>Text</p></text>
        </text-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "<p>Text</p>"
        assert val.value_type == KnoraValueType.RICHTEXT_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_text_simpletext_value(self):
        xml_val = etree.fromstring("""
        <text-prop name=":hasProp">
            <text encoding="utf8">Text</text>
        </text-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "Text"
        assert val.value_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_text_simpletext_value_no_text(self):
        xml_val = etree.fromstring("""
        <text-prop name=":hasProp">
            <text encoding="utf8"></text>
        </text-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert not val.value
        assert val.value_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_time_value(self):
        xml_val = etree.fromstring("""
        <time-prop name=":hasProp">
            <time>2019-10-23T13:45:12.01-14:00</time>
        </time-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "2019-10-23T13:45:12.01-14:00"
        assert val.value_type == KnoraValueType.TIME_VALUE
        assert not val.permissions_id
        assert not val.comment

    def test_uri_value(self):
        xml_val = etree.fromstring("""
        <uri-prop name=":hasProp">
            <uri>https://dasch.swiss</uri>
        </uri-prop>
        """)
        result = _parse_one_value(xml_val, IRI_LOOKUP)
        assert len(result) == 1
        val = result.pop(0)
        assert val.prop_name == HAS_PROP
        assert val.value == "https://dasch.swiss"
        assert val.value_type == KnoraValueType.URI_VALUE
        assert not val.permissions_id
        assert not val.comment


class TestParseFileValues:
    def test_iiif_no_legal_info(self):
        xml_val = etree.fromstring("""
        <iiif-uri>https://iiif.uri/full.jpg</iiif-uri>
        """)
        val = _parse_iiif_uri(xml_val)
        assert val.value == "https://iiif.uri/full.jpg"
        assert val.value_type == KnoraValueType.STILL_IMAGE_IIIF
        assert not val.metadata.license_iri
        assert not val.metadata.copyright_holder
        assert not val.metadata.authorship_id
        assert not val.metadata.permissions_id

    def test_iiif_with_legal_info(self):
        xml_val = etree.fromstring("""
        <iiif-uri license="license_iri"
                  copyright-holder="copy"
                  authorship-id="auth"
        >https://iiif.uri/full.jpg</iiif-uri>
        """)
        val = _parse_iiif_uri(xml_val)
        assert val.value == "https://iiif.uri/full.jpg"
        assert val.value_type == KnoraValueType.STILL_IMAGE_IIIF
        assert val.metadata.license_iri == "license_iri"
        assert val.metadata.copyright_holder == "copy"
        assert val.metadata.authorship_id == "auth"
        assert not val.metadata.permissions_id

    def test_bitstream_with_permissions(self):
        xml_val = etree.fromstring("""
        <bitstream permissions="open">this/is/filepath/file.z</bitstream>
        """)
        val = _parse_file_values(xml_val)
        assert val.value == "this/is/filepath/file.z"
        assert val.value_type == KnoraValueType.ARCHIVE_FILE
        assert not val.metadata.license_iri
        assert not val.metadata.copyright_holder
        assert not val.metadata.authorship_id
        assert val.metadata.permissions_id == "open"

    def test_bitstream_with_legal_info(self):
        xml_val = etree.fromstring("""
        <bitstream license="http://rdfh.ch/licenses/unknown"
                   copyright-holder="DaSCH"
                   authorship-id="authorship_1"
        >this/is/filepath/file.z</bitstream>
        """)
        val = _parse_file_values(xml_val)
        assert val.value == "this/is/filepath/file.z"
        assert val.value_type == KnoraValueType.ARCHIVE_FILE
        assert val.metadata.license_iri == "http://rdfh.ch/licenses/unknown"
        assert val.metadata.copyright_holder == "DaSCH"
        assert val.metadata.authorship_id == "authorship_1"
        assert not val.metadata.permissions_id


class TestFileTypeInfo:
    @pytest.mark.parametrize(
        "file_name", ["test.zip", "test.tar", "test.gz", "test.z", "test.tgz", "test.gzip", "test.7z"]
    )
    def test_archive(self, file_name: str):
        assert _get_file_value_type(file_name) == KnoraValueType.ARCHIVE_FILE

    @pytest.mark.parametrize("file_name", ["test.mp3", "test.wav"])
    def test_audio(self, file_name: str):
        assert _get_file_value_type(file_name) == KnoraValueType.AUDIO_FILE

    @pytest.mark.parametrize(
        "file_name", ["test.pdf", "test.DOC", "test.docx", "test.xls", "test.xlsx", "test.ppt", "test.pptx"]
    )
    def test_document(self, file_name: str):
        assert _get_file_value_type(file_name) == KnoraValueType.DOCUMENT_FILE

    def test_moving_image(self):
        assert _get_file_value_type("test.mp4") == KnoraValueType.MOVING_IMAGE_FILE

    @pytest.mark.parametrize(
        "file_name", ["test.jpg", "test.jpeg", "path/test.jp2", "test.png", "test.tif", "test.tiff", "test.jpx"]
    )
    def test_still_image(self, file_name: str):
        assert _get_file_value_type(file_name) == KnoraValueType.STILL_IMAGE_FILE

    @pytest.mark.parametrize(
        "file_name",
        ["path/test.odd", "test.rng", "test.txt", "test.xml", "test.xsd", "test.xsl", "test.csv", "test.json"],
    )
    def test_text(self, file_name: str):
        assert _get_file_value_type(file_name) == KnoraValueType.TEXT_FILE

    def test_unknown(self):
        assert not _get_file_value_type("file.unknown")

    def test_none(self):
        assert not _get_file_value_type(None)


def test_create_from_local_name_to_absolute_iri_lookup(root_with_2_resources):
    result = _create_from_local_name_to_absolute_iri_lookup(root_with_2_resources, API_URL)
    expected = {
        ":minimalResource": f"{DEFAULT_ONTO_NAMESPACE}minimalResource",
        "hasLinkTo": f"{KNORA_API_STR}hasLinkTo",
    }
    assert result == expected


@pytest.mark.parametrize(
    ("local_name", "expected"),
    [
        (":defaultOnto", f"{DEFAULT_ONTO_NAMESPACE}defaultOnto"),
        ("knora-api:localName", f"{KNORA_API_STR}localName"),
        ("knoraApiNoPrefix", f"{KNORA_API_STR}knoraApiNoPrefix"),
        ("other-onto:localName", f"{API_URL}/ontology/0000/other-onto/v2#localName"),
        ("default:withDefaultOnto", f"{DEFAULT_ONTO_NAMESPACE}withDefaultOnto"),
    ],
)
def test_get_one_absolute_iri(local_name, expected):
    result = _get_one_absolute_iri(local_name, "0000", "default", API_URL)
    assert result == expected



import json
from json import JSONDecodeError
from pathlib import Path

import regex
from lxml import etree

from dsp_tools.commands.validate_data.constants import AUDIO_SEGMENT_RESOURCE
from dsp_tools.commands.validate_data.constants import VIDEO_SEGMENT_RESOURCE
from dsp_tools.commands.validate_data.mappers import SEGMENT_TAG_TO_PROP_MAPPER
from dsp_tools.commands.validate_data.mappers import XML_ATTRIB_TO_PROP_TYPE_MAPPER
from dsp_tools.commands.validate_data.mappers import XML_TAG_TO_VALUE_TYPE_MAPPER
from dsp_tools.error.exceptions import BaseError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.data_deserialised import DataDeserialised
from dsp_tools.utils.xml_parsing.models.data_deserialised import MigrationMetadata
from dsp_tools.utils.xml_parsing.models.data_deserialised import PropertyObject
from dsp_tools.utils.xml_parsing.models.data_deserialised import ResourceDeserialised
from dsp_tools.utils.xml_parsing.models.data_deserialised import TripleObjectType
from dsp_tools.utils.xml_parsing.models.data_deserialised import TriplePropertyType
from dsp_tools.utils.xml_parsing.models.data_deserialised import ValueInformation
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType


def get_data_deserialised(root: etree._Element) -> tuple[str, DataDeserialised]:
    """
    Takes the root of an XML
    Extracts the data of the project and transforms all its resources.

    Args:
        root: root of an xml with qnames and comments removed

    Returns:
        Shortcode and deserialised data
    """
    shortcode = root.attrib["shortcode"]
    data_deserialised = _deserialise_all_resources(root)
    return shortcode, data_deserialised


def _deserialise_all_resources(root: etree._Element) -> DataDeserialised:
    all_res: list[ResourceDeserialised] = []
    for res in root.iterdescendants(tag="resource"):
        all_res.append(_deserialise_one_resource(res))
    return DataDeserialised(all_res)


def _extract_metadata(element: etree._Element) -> list[PropertyObject]:
    property_objects = []
    for k, v in element.attrib.items():
        if not (knora_prop := XML_ATTRIB_TO_PROP_TYPE_MAPPER.get(k)):
            continue
        object_type = TripleObjectType.STRING
        if knora_prop == TriplePropertyType.KNORA_LICENSE:
            object_type = TripleObjectType.IRI
        property_objects.append(PropertyObject(property_type=knora_prop, object_value=v, object_type=object_type))
    return property_objects


def _deserialise_one_resource(resource: etree._Element) -> ResourceDeserialised:
    res_type = resource.attrib["restype"]
    asset_value = None
    if res_type in {VIDEO_SEGMENT_RESOURCE, AUDIO_SEGMENT_RESOURCE}:
        values = _deserialise_segment_properties(resource)
    else:
        values, asset_value = _deserialise_generic_properties(resource)
    metadata = _extract_metadata(resource)
    metadata.extend(_get_all_stand_off_links(values))
    metadata.append(PropertyObject(TriplePropertyType.RDFS_LABEL, resource.attrib["label"], TripleObjectType.STRING))
    metadata.append(PropertyObject(TriplePropertyType.RDF_TYPE, res_type, TripleObjectType.IRI))
    return ResourceDeserialised(
        res_id=resource.attrib["id"],
        property_objects=metadata,
        values=values,
        asset_value=asset_value,
        migration_metadata=_deserialise_migration_metadata(resource),
    )


def _deserialise_migration_metadata(resource: etree._Element) -> MigrationMetadata:
    date = resource.attrib.get("creation_date")
    creation_date = DateTimeStamp(date) if date else None
    return MigrationMetadata(
        iri=resource.attrib.get("iri"),
        ark=resource.attrib.get("ark"),
        creation_date=creation_date,
    )


def _deserialise_generic_properties(resource: etree._Element) -> tuple[list[ValueInformation], ValueInformation | None]:
    values = []
    asset_value = None
    for val in resource.iterchildren():
        if val.tag == "bitstream":
            asset_value = _deserialise_bitstream(val)
        else:
            # iiif links are handled like all other values
            values.extend(_deserialise_one_property(val))
    return values, asset_value


def _deserialise_one_property(prop_ele: etree._Element) -> list[ValueInformation]:
    match prop_ele.tag:
        case (
            (
                "boolean-prop"
                | "color-prop"
                | "date-prop"
                | "decimal-prop"
                | "geoname-prop"
                | "integer-prop"
                | "resptr-prop"
                | "time-prop"
                | "uri-prop"
            ) as prop_tag
        ):
            return _extract_generic_value_information(prop_ele, XML_TAG_TO_VALUE_TYPE_MAPPER[prop_tag])
        case "list-prop":
            return _extract_list_value_information(prop_ele)
        case "text-prop":
            return _extract_text_value_information(prop_ele)
        case "iiif-uri":
            return _deserialise_iiif_uri(prop_ele)
        case "geometry-prop":
            return _extract_geometry_value_information(prop_ele)
        case _:
            return []


def _extract_generic_value_information(prop: etree._Element, value_type: KnoraValueType) -> list[ValueInformation]:
    prop_name = prop.attrib["name"]
    return [
        ValueInformation(
            user_facing_prop=prop_name,
            user_facing_value=val.text,
            knora_type=value_type,
            value_metadata=_extract_metadata(val),
        )
        for val in prop.iterchildren()
    ]


def _extract_list_value_information(prop: etree._Element) -> list[ValueInformation]:
    prop_name = prop.attrib["name"]
    list_name = prop.attrib["list"]
    all_vals = []
    for val in prop.iterchildren():
        found_value = f"{list_name} / {val.text}" if val.text else None
        all_vals.append(
            ValueInformation(
                user_facing_prop=prop_name,
                user_facing_value=found_value,
                knora_type=KnoraValueType.LIST_VALUE,
                value_metadata=_extract_metadata(val),
            )
        )
    return all_vals


def _extract_text_value_information(prop: etree._Element) -> list[ValueInformation]:
    prop_name = prop.attrib["name"]
    all_vals = []
    for val in prop.iterchildren():
        encoding = val.attrib["encoding"]
        match encoding:
            case "utf8":
                val_type = KnoraValueType.SIMPLETEXT_VALUE
            case "xml":
                val_type = KnoraValueType.RICHTEXT_VALUE
            case _:
                raise BaseError(f"Unknown encoding: {encoding}.")
        all_vals.append(
            ValueInformation(
                user_facing_prop=prop_name,
                user_facing_value=_get_text_as_string(val),
                knora_type=val_type,
                value_metadata=_extract_metadata(val),
            )
        )
    return all_vals


def _get_text_as_string(value: etree._Element) -> str | None:
    if len(value):
        text_list = []
        if found := value.text:
            text_list = [found]
        text_list.extend([etree.tostring(child, encoding="unicode", method="xml") for child in value])
        return "".join(text_list).strip()
    else:
        return value.text


def _get_all_stand_off_links(values: list[ValueInformation]) -> list[PropertyObject]:
    stand_offs = []
    for val in values:
        if val.knora_type.RICHTEXT_VALUE:
            stand_offs.extend(_get_stand_off_links(val.user_facing_value))
    return stand_offs


def _get_stand_off_links(text: str | None) -> list[PropertyObject]:
    if not text:
        return []
    links = set(regex.findall(pattern='href="IRI:(.*?):IRI"', string=text))
    return [PropertyObject(TriplePropertyType.KNORA_STANDOFF_LINK, lnk, TripleObjectType.IRI) for lnk in links]


def _extract_geometry_value_information(prop: etree._Element) -> list[ValueInformation]:
    prop_name = prop.attrib["name"]

    def check_for_geometry_json(value: str | None) -> str | None:
        if not value:
            return None
        try:
            return json.dumps(json.loads(value))
        except JSONDecodeError:
            return None

    return [
        ValueInformation(
            user_facing_prop=prop_name,
            user_facing_value=check_for_geometry_json(val.text),
            knora_type=KnoraValueType.GEOM_VALUE,
            value_metadata=_extract_metadata(val),
        )
        for val in prop.iterchildren()
    ]


def _deserialise_iiif_uri(value: etree._Element) -> list[ValueInformation]:
    if (file_str := value.text) is None:
        return []
    file_str = file_str.strip()
    return [
        ValueInformation(
            user_facing_prop=f"{KNORA_API_STR}hasStillImageFileValue",
            user_facing_value=file_str,
            knora_type=KnoraValueType.STILL_IMAGE_IIIF,
            value_metadata=_extract_metadata(value),
        )
    ]


def _deserialise_bitstream(value: etree._Element) -> ValueInformation | None:
    if (file_str := value.text) is None:
        return None
    file_str = file_str.strip()
    file_type, prop_str = _get_file_value_type(file_str)
    if not file_type:
        return None
    return ValueInformation(
        user_facing_prop=prop_str,
        user_facing_value=file_str,
        knora_type=file_type,
        value_metadata=_extract_metadata(value),
    )


def _get_file_value_type(file_name: str) -> tuple[KnoraValueType | None, str]:  # noqa:PLR0911 (Too many return statements)
    file_ending = Path(file_name).suffix[1:].lower()
    match file_ending:
        case "zip" | "tar" | "gz" | "z" | "tgz" | "gzip" | "7z":
            return KnoraValueType.ARCHIVE_FILE, f"{KNORA_API_STR}hasArchiveFileValue"
        case "mp3" | "wav":
            return KnoraValueType.AUDIO_FILE, f"{KNORA_API_STR}hasAudioFileValue"
        case "pdf" | "doc" | "docx" | "xls" | "xlsx" | "ppt" | "pptx":
            return KnoraValueType.DOCUMENT_FILE, f"{KNORA_API_STR}hasDocumentFileValue"
        case "mp4":
            return KnoraValueType.MOVING_IMAGE_FILE, f"{KNORA_API_STR}hasMovingImageFileValue"
        # jpx is the extension of the files returned by dsp-ingest
        case "jpg" | "jpeg" | "jp2" | "png" | "tif" | "tiff" | "jpx":
            return KnoraValueType.STILL_IMAGE_FILE, f"{KNORA_API_STR}hasStillImageFileValue"
        case "odd" | "rng" | "txt" | "xml" | "xsd" | "xsl" | "csv" | "json":
            return KnoraValueType.TEXT_FILE, f"{KNORA_API_STR}hasTextFileValue"
        case _:
            return None, ""


def _deserialise_segment_properties(resource: etree._Element) -> list[ValueInformation]:
    values = []
    for val in resource.iterchildren():
        prop_name = str(val.tag)
        property_objects = _extract_metadata(val)
        if prop_name == "hasSegmentBounds":
            property_objects.append(
                PropertyObject(
                    TriplePropertyType.KNORA_INTERVAL_START,
                    val.attrib["segment_start"],
                    TripleObjectType.DECIMAL,
                )
            )
            property_objects.append(
                PropertyObject(
                    TriplePropertyType.KNORA_INTERVAL_END,
                    val.attrib["segment_end"],
                    TripleObjectType.DECIMAL,
                )
            )
        values.append(
            ValueInformation(
                user_facing_prop=f"{KNORA_API_STR}{prop_name}",
                user_facing_value=_get_text_as_string(val),
                knora_type=SEGMENT_TAG_TO_PROP_MAPPER[prop_name],
                value_metadata=property_objects,
            )
        )
    return values



KNOWN_XML_TAGS = [  # defined at https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/
    "a( [^>]+)?",  # <a> has attributes
    "footnote( [^>]+)?",  # the footnote text is in the attributes
    "p",
    "em",
    "strong",
    "u",
    "sub",
    "sup",
    "strike",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "ol",
    "ul",
    "li",
    "tbody",
    "table",
    "tr",
    "td",
    "br",
    "hr",
    "pre",
    "cite",
    "blockquote",
    "code",
]
XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import auto


@dataclass
class ParsedResource:
    res_id: str
    res_type: str
    label: str
    permissions_id: str | None
    values: list[ParsedValue]
    file_value: ParsedFileValue | None
    migration_metadata: ParsedMigrationMetadata | None


@dataclass
class ParsedMigrationMetadata:
    iri: str | None
    ark: str | None
    creation_date: str | None


@dataclass
class ParsedValue:
    prop_name: str
    value: str | tuple[str | None, str | None] | None
    value_type: KnoraValueType
    permissions_id: str | None
    comment: str | None


@dataclass
class ParsedFileValue:
    value: str | None
    value_type: KnoraValueType | None
    metadata: ParsedFileValueMetadata


@dataclass
class ParsedFileValueMetadata:
    license_iri: str | None
    copyright_holder: str | None
    authorship_id: str | None
    permissions_id: str | None


class KnoraValueType(Enum):
    """
    Maps to a knora value type, for example: BOOLEAN_VALUE -> knora-api:BooleanValue
    """

    BOOLEAN_VALUE = auto()
    COLOR_VALUE = auto()
    DATE_VALUE = auto()
    DECIMAL_VALUE = auto()
    GEONAME_VALUE = auto()
    GEOM_VALUE = auto()
    INT_VALUE = auto()
    INTERVAL_VALUE = auto()
    LINK_VALUE = auto()
    LIST_VALUE = auto()
    SIMPLETEXT_VALUE = auto()
    RICHTEXT_VALUE = auto()
    TIME_VALUE = auto()
    URI_VALUE = auto()

    ARCHIVE_FILE = auto()
    AUDIO_FILE = auto()
    DOCUMENT_FILE = auto()
    MOVING_IMAGE_FILE = auto()
    STILL_IMAGE_FILE = auto()
    STILL_IMAGE_IIIF = auto()
    TEXT_FILE = auto()



from pathlib import Path
from typing import Iterator
from typing import Never
from typing import assert_never
from typing import cast

import pytest
from rdflib import BNode
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.get_user_validation_message import _filter_out_duplicate_problems
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.query_validation_result import _extract_base_info_of_resource_results
from dsp_tools.commands.validate_data.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _get_parsed_graphs
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from dsp_tools.error.custom_warnings import DspToolsUserInfo
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts
from test.e2e.setup_testcontainers.setup import get_containers


@pytest.fixture(scope="module")
def container_ports() -> Iterator[ExternalContainerPorts]:
    with get_containers() as metadata:
        yield metadata.ports


@pytest.fixture(scope="module")
def creds(container_ports: ExternalContainerPorts) -> ServerCredentials:
    return ServerCredentials(
        "root@example.com",
        "test",
        f"http://0.0.0.0:{container_ports.api}",
        f"http://0.0.0.0:{container_ports.ingest}",
    )


@pytest.fixture(scope="module")
def _create_projects(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/validate-data/generic/project.json"), creds)


@pytest.fixture(scope="module")
def api_url(container_ports: ExternalContainerPorts) -> str:
    return f"http://0.0.0.0:{container_ports.api}"


@pytest.fixture(scope="module")
def shacl_validator(api_url: str) -> ShaclValidator:
    return ShaclValidator(api_url)


@pytest.fixture(scope="module")
def unique_value_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/unique_value_violation.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def file_value_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/file_value_violation.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def dsp_inbuilt_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_violation.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def cardinality_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/cardinality_violation.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def content_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/content_violation.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def value_type_violation(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/value_type_violation.xml")
    match = r"Angular brackets in the format of <text> were found in text properties with encoding=utf8"
    with pytest.warns(DspToolsUserInfo, match=match):
        graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


@pytest.fixture(scope="module")
def every_violation_combination_once(
    _create_projects: Iterator[None], api_url: str, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/every_violation_combination_once.xml")
    graphs = _get_parsed_graphs(api_url, file)
    return _get_validation_result(graphs, shacl_validator, None)


def test_cardinality_violation(cardinality_violation: ValidationReportGraphs) -> None:
    assert not cardinality_violation.conforms


def test_reformat_cardinality_violation(cardinality_violation: ValidationReportGraphs) -> None:
    result = reformat_validation_graph(cardinality_violation)
    expected_info_tuples = [
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_min_card", ProblemType.MIN_CARD),
        ("super_prop_no_card", ProblemType.NON_EXISTING_CARD),
    ]
    assert not result.unexpected_results
    assert len(result.problems) == len(expected_info_tuples)
    sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
    for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.problem_type == expected_info[1]


def test_content_violation(content_violation: ValidationReportGraphs) -> None:
    assert not content_violation.conforms


def test_reformat_content_violation(content_violation: ValidationReportGraphs) -> None:
    result = reformat_validation_graph(content_violation)
    assert not result.unexpected_results
    sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
    expected_info_tuples = [
        (
            "comment_on_value_empty",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        (
            "comment_on_value_whitespace",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        ("empty_label", "rdfs:label", "The label must be a non-empty string"),
        ("empty_text_rich", "onto:testRichtext", "The value must be a non-empty string"),
        ("empty_text_simple", "onto:testTextarea", "The value must be a non-empty string"),
        ("geoname_not_number", "onto:testGeoname", "The value must be a valid geoname code"),
        ("link_target_non_existent", "onto:testHasLinkTo", "other"),
        ("link_target_wrong_class", "onto:testHasLinkToCardOneResource", "id_9_target"),
        (
            "list_name_attrib_empty",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_name_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_node_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "richtext_standoff_link_nonexistent",
            "hasStandoffLinkTo",
            "A stand-off link must target an existing resource.",
        ),
        ("text_only_whitespace_simple", "onto:testTextarea", "The value must be a non-empty string"),
    ]
    assert len(result.problems) == len(expected_info_tuples)
    for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.prop_name == expected_info[1]
        if one_result.problem_type == ProblemType.INPUT_REGEX:
            assert one_result.expected == expected_info[2]
        elif one_result.problem_type == ProblemType.GENERIC:
            assert one_result.message == expected_info[2]
        elif one_result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH:
            assert one_result.input_value == expected_info[2]
        elif one_result.problem_type == ProblemType.INEXISTENT_LINKED_RESOURCE:
            assert one_result.input_value == expected_info[2]
        else:
            nev: Never = cast(Never, one_result.problem_type)
            assert_never(nev)


def test_value_type_violation(value_type_violation: ValidationReportGraphs) -> None:
    assert not value_type_violation.conforms


def test_reformat_value_type_violation(value_type_violation: ValidationReportGraphs) -> None:
    result = reformat_validation_graph(value_type_violation)
    assert not result.unexpected_results
    # "is_link_should_be_text" gives two types of validation errors, this function removes the duplicates
    filtered_problems = _filter_out_duplicate_problems(result.problems)
    flattened_problems = [item for sublist in filtered_problems for item in sublist]
    sorted_problems = sorted(flattened_problems, key=lambda x: x.res_id)
    expected_info_tuples = [
        ("bool_wrong_value_type", "This property requires a BooleanValue", "onto:testBoolean"),
        ("color_wrong_value_type", "This property requires a ColorValue", "onto:testColor"),
        ("date_wrong_value_type", "This property requires a DateValue", "onto:testSubDate1"),
        ("decimal_wrong_value_type", "This property requires a DecimalValue", "onto:testDecimalSimpleText"),
        ("geoname_wrong_value_type", "This property requires a GeonameValue", "onto:testGeoname"),
        ("integer_wrong_value_type", "This property requires a IntValue", "onto:testIntegerSimpleText"),
        ("is_date_should_be_simpletext", "This property requires a TextValue", "onto:testTextarea"),
        ("is_link_should_be_integer", "This property requires a IntValue", "onto:testIntegerSpinbox"),
        ("is_link_should_be_text", "TextValue without formatting", "onto:testTextarea"),
        ("link_wrong_value_type", "This property requires a LinkValue", "onto:testHasLinkTo"),
        ("list_wrong_value_type", "This property requires a ListValue", "onto:testListProp"),
        ("richtext_wrong_value_type", "TextValue with formatting", "onto:testRichtext"),
        ("simpletext_wrong_value_type", "TextValue without formatting", "onto:testTextarea"),
        ("time_wrong_value_type", "This property requires a TimeValue", "onto:testTimeValue"),
        ("uri_wrong_value_type", "This property requires a UriValue", "onto:testUriValue"),
    ]
    assert len(sorted_problems) == len(expected_info_tuples)
    for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
        assert one_result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert one_result.res_id == expected_info[0]
        assert one_result.expected == expected_info[1]
        assert one_result.prop_name == expected_info[2]


def test_unique_value_violation(unique_value_violation: ValidationReportGraphs) -> None:
    assert not unique_value_violation.conforms


def test_file_value_cardinality_violation(file_value_violation: ValidationReportGraphs) -> None:
    assert not file_value_violation.conforms


def test_dsp_inbuilt_violation(dsp_inbuilt_violation: ValidationReportGraphs) -> None:
    assert not dsp_inbuilt_violation.conforms


class TestReformatValidationGraph:
    def test_reformat_unique_value_violation(self, unique_value_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(unique_value_violation)
        expected_ids = [
            "identical_values_LinkValue",
            "identical_values_listNode",
            "identical_values_valueAsString",
            "identical_values_valueHas",
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_ids)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_id in zip(sorted_problems, expected_ids):
            assert one_result.problem_type == ProblemType.DUPLICATE_VALUE
            assert one_result.res_id == expected_id

    def test_reformat_file_value_violation(self, file_value_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(file_value_violation)
        expected_info_tuples = [
            # each type of missing legal info (authorship, copyright, license) produces one violation
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("empty_license", ProblemType.GENERIC),
            ("id_archive_missing", ProblemType.FILE_VALUE),
            ("id_archive_unknown", ProblemType.FILE_VALUE),
            ("id_audio_missing", ProblemType.FILE_VALUE),
            ("id_audio_unknown", ProblemType.FILE_VALUE),
            ("id_document_missing", ProblemType.FILE_VALUE),
            ("id_document_unknown", ProblemType.FILE_VALUE),
            ("id_resource_without_representation", ProblemType.FILE_VALUE_PROHIBITED),
            ("id_still_image_missing", ProblemType.FILE_VALUE),
            ("id_still_image_unknown", ProblemType.FILE_VALUE),
            ("id_text_missing", ProblemType.FILE_VALUE),
            ("id_text_unknown", ProblemType.FILE_VALUE),
            ("id_video_missing", ProblemType.FILE_VALUE),
            ("id_video_unknown", ProblemType.FILE_VALUE),
            ("id_wrong_file_type", ProblemType.FILE_VALUE),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("iiif_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("inexistent_license_iri", ProblemType.GENERIC),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]

    def test_reformat_dsp_inbuilt_violation(self, dsp_inbuilt_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(dsp_inbuilt_violation)
        expected_info_tuples = [
            ("audio_segment_target_is_video", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("audio_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("link_obj_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("missing_isPartOf", ProblemType.GENERIC),
            ("missing_seqnum", ProblemType.GENERIC),
            ("region_invalid_geometry", ProblemType.INPUT_REGEX),
            ("region_isRegionOf_resource_does_not_exist", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("region_isRegionOf_resource_not_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("target_must_be_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("target_must_be_an_image_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("video_segment_start_larger_than_end", ProblemType.GENERIC),
            ("video_segment_target_is_audio", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("video_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert one_result.problem_type == expected_info[1]
            assert one_result.res_id == expected_info[0]


def test_extract_identifiers_of_resource_results(every_violation_combination_once: ValidationReportGraphs) -> None:
    report_and_onto = every_violation_combination_once.validation_graph + every_violation_combination_once.onto_graph
    data_and_onto = every_violation_combination_once.data_graph + every_violation_combination_once.onto_graph
    result = _extract_base_info_of_resource_results(report_and_onto, data_and_onto)
    result_sorted = sorted(result, key=lambda x: str(x.focus_node_iri))
    expected_iris = [
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/bitstream_no_legal_info"), None),
        (URIRef("http://data/empty_label"), None),
        (URIRef("http://data/geoname_not_number"), None),
        (URIRef("http://data/id_card_one"), None),
        (URIRef("http://data/id_closed_constraint"), None),
        (URIRef("http://data/id_max_card"), None),
        (URIRef("http://data/id_missing_file_value"), None),
        (URIRef("http://data/identical_values"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/image_no_legal_info"), None),
        (URIRef("http://data/inexistent_license_iri"), None),
        (URIRef("http://data/link_target_non_existent"), BNode),
        (URIRef("http://data/link_target_wrong_class"), BNode),
        (URIRef("http://data/list_node_non_existent"), BNode),
        (URIRef("http://data/missing_seqnum"), None),
        (URIRef("http://data/richtext_standoff_link_nonexistent"), None),
        (URIRef("http://data/simpletext_wrong_value_type"), BNode),
        (URIRef("http://data/uri_wrong_value_type"), None),
        (URIRef("http://data/video_segment_start_larger_than_end"), None),
        (URIRef("http://data/video_segment_wrong_bounds"), None),
        (URIRef("http://data/video_segment_wrong_bounds"), None),
    ]
    assert len(result) == len(expected_iris)
    for result_info, expected_iri in zip(result_sorted, expected_iris):
        assert result_info.focus_node_iri == expected_iri[0]
        if expected_iri[1] is None:
            assert not result_info.detail
        else:
            detail_base_info = result_info.detail
            assert isinstance(detail_base_info, DetailBaseInfo)
            assert isinstance(detail_base_info.detail_bn, expected_iri[1])


def test_every_violation_combination_once(every_violation_combination_once: ValidationReportGraphs) -> None:
    assert not every_violation_combination_once.conforms


def test_reformat_every_constraint_once(every_violation_combination_once: ValidationReportGraphs) -> None:
    result = reformat_validation_graph(every_violation_combination_once)
    expected_info_tuples = [
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("bitstream_no_legal_info", ProblemType.GENERIC),
        ("empty_label", ProblemType.INPUT_REGEX),
        ("geoname_not_number", ProblemType.INPUT_REGEX),
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_missing_file_value", ProblemType.FILE_VALUE),
        ("identical_values", ProblemType.DUPLICATE_VALUE),
        ("image_no_legal_info", ProblemType.GENERIC),
        ("image_no_legal_info", ProblemType.GENERIC),
        ("image_no_legal_info", ProblemType.GENERIC),
        ("inexistent_license_iri", ProblemType.GENERIC),
        ("link_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("link_target_wrong_class", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("list_node_non_existent", ProblemType.GENERIC),
        ("missing_seqnum", ProblemType.GENERIC),
        ("richtext_standoff_link_nonexistent", ProblemType.GENERIC),
        ("simpletext_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
        ("uri_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
        ("video_segment_start_larger_than_end", ProblemType.GENERIC),
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
    ]
    assert not result.unexpected_results
    sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
    assert len(result.problems) == len(expected_info_tuples)
    for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.problem_type == expected_info[1]


if __name__ == "__main__":
    pytest.main([__file__])
